# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import json
from inspect import isclass

from kqcircuits.defaults import default_layers, default_faces, default_parameter_values
from kqcircuits.pya_resolver import pya, is_standalone_session
from kqcircuits.util.geometry_helper import get_cell_path_length
from kqcircuits.util.library_helper import load_libraries, to_library_name, to_module_name, element_by_class_name
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.util.refpoints import Refpoints, WaveguideToSimPort


def get_refpoints(layer, cell, cell_transf=pya.DTrans(), rec_levels=None):
    """Returns Refpoints object for extracting reference points from given layer and cell.

    Args:
        layer: layer specification for source of refpoints
        cell: cell containing the refpoints
        cell_transf: transform for converting refpoints into target coordinate system
        rec_levels: recursion level when looking for refpoints from subcells. Set to 0 to disable recursion.

    Returns:
        Refpoints object, which behaves like dictionary, where keys are refpoints names, values are DPoints.

    """
    return Refpoints(layer, cell, cell_transf, rec_levels)


def insert_cell_into(target_cell, cell, trans=None, inst_name=None, label_trans=None, align_to=None, align=None,
                     rec_levels=0, **parameters):
    """Inserts a subcell into a given target cell.

    Note: This general method is useful to insert cells or elements into a static cell. To insert cells into an
    ``Element``, use the elements' ``insert_cell`` method, which has additional features such as parameter inheritance.

    It will use the given ``cell`` object or if ``cell`` is an Element class' name then directly
    take the provided keyword arguments to first create the cell object.

    If `inst_name` given, a label ``inst_name`` is added to labels layer at the ``base`` refpoint and `label_trans`
    transformation.

    Arguments:
        target_cell: Cell object to insert into
        cell: cell object or Element class name
        trans: used transformation for placement. None by default, which places the subcell into the coordinate
            origin of the parent cell. If `align` and `align_to` arguments are used, `trans` is applied to the
            `cell` before alignment transform which allows for example rotation of the `cell` before placement.
        inst_name: possible instance name inserted into subcell properties under `id`. Default is None
        label_trans: relative transformation for the instance name label
        align_to: ``DPoint`` or ``DVector`` location in parent cell coordinates for alignment of cell. Default is None
        align: name of the ``cell`` refpoint aligned to argument ``align_to``. Default is None
        rec_levels: recursion level when looking for refpoints from subcells. Set to 0 to disable recursion.
        **parameters: PCell parameters for the element, as keyword argument

    Return:
        tuple of placed cell instance and reference points with the same transformation
    """
    layout = target_cell.layout()
    if isclass(cell):
        cell = cell.create(layout, **parameters)

    if trans is None:
        trans = pya.DTrans()
    if (align_to and align) is not None:
        align = get_refpoints(layout.layer(default_layers["refpoints"]), cell, trans, rec_levels=rec_levels)[align]
        trans = pya.DCplxTrans(align_to - align) * trans

    cell_inst = target_cell.insert(pya.DCellInstArray(cell.cell_index(), trans))

    refpoints_abs = get_refpoints(layout.layer(default_layers["refpoints"]), cell, cell_inst.dcplx_trans, rec_levels)
    if inst_name is not None:
        cell_inst.set_property("id", inst_name)
        if label_trans is not None:
            label_trans_str = pya.DCplxTrans(label_trans).to_s()  # must be saved as string to avoid errors
            cell_inst.set_property("label_trans", label_trans_str)
    return cell_inst, refpoints_abs

class Element(pya.PCellDeclarationHelper):
    """Element PCell declaration.

    PCell parameters for an element are defined as class attributes of Param type.
    Elements have ports.
    """

    LIBRARY_NAME = "Element Library"
    LIBRARY_DESCRIPTION = "Superconducting quantum circuit library for elements."
    LIBRARY_PATH = "elements"

    a = Param(pdt.TypeDouble, "Width of center conductor", 10, unit="μm")
    b = Param(pdt.TypeDouble, "Width of gap", 6, unit="μm")
    n = Param(pdt.TypeInt, "Number of points on turns", 64)
    r = Param(pdt.TypeDouble, "Turn radius", 100, unit="μm")
    margin = Param(pdt.TypeDouble, "Margin of the protection layer", 5, unit="μm")
    face_ids = Param(pdt.TypeList, "Chip face IDs list", ["1t1", "2b1", "1b1", "2t1"])
    display_name = Param(pdt.TypeString, "Name displayed in GUI (empty for default)", "")
    protect_opposite_face = Param(pdt.TypeBoolean, "Add opposite face protection too", False)
    opposing_face_id_groups = Param(pdt.TypeList, "Opposing face ID groups (list of lists)", [["1t1", "2b1"]],
                                    hidden=True)

    def __init__(self):
        ""
        super().__init__()

        cls = type(self)
        mro = cls.__mro__

        # We may need to redefine a Param object, because multiple classes may refer to the same Param object
        # due to inheritance, so modifying the existing Param object could affect other classes.
        def _redef_param(p, v, **kwargs):
            np = Param(p.data_type, p.description, v, **{**p.kwargs, **kwargs})
            np.__set_name__(cls, p.name)
            setattr(type(self), p.name, np)
            return np

        # Set and hide *_type parameter in classes inheriting from a * abstract class
        base = cls._get_abstract()
        if hasattr(base, "default_type") and getattr(base, "build") == getattr(Element, "build"):
            params = Param.get_all(base)
            mod = to_module_name(base.__name__)
            if f"{mod}_type" in params:
                subtype = to_library_name(cls.__name__)
                _redef_param(params[f"{mod}_type"], subtype, choices=[subtype], hidden=True)

        # create KLayout's PCellParameterDeclaration objects
        self._param_value_map = {}
        for name, p in cls.get_schema().items():
            self._param_value_map[name] = len(self._param_decls)
            # Override default value based on default_parameter_values if needed.
            for cl in mro:
                cls_name = cl.__qualname__
                if cls_name in default_parameter_values and name in default_parameter_values[cls_name]:
                    # Ensure that the `cl` default overrides the value only if it is not overridden
                    # by another class below `cl` in the hierarchy.
                    if cl != cls and cl.__dict__[name] != p:
                        break
                    p = _redef_param(p, default_parameter_values[cls_name][name])
                    break
            self._add_parameter(name, p.data_type, p.description, default=p.default, **p.kwargs)

    @staticmethod
    def create_cell_from_shape(layout, name):
        load_libraries(path=Element.LIBRARY_PATH)
        return layout.create_cell(name, Element.LIBRARY_NAME)

    @classmethod
    def create(cls, layout, library=None, **parameters) -> pya.Cell:
        """Create cell for this element in layout.

        Args:
            layout: pya.Layout object where this cell is created
            library: LIBRARY_NAME of the calling PCell instance
            **parameters: PCell parameters for the element as keyword arguments
        """
        cell = Element._create_cell(cls, layout, library, **parameters)
        setattr(cell, "length", lambda: get_cell_path_length(cell))
        return cell

    @classmethod
    def create_subtype(cls, layout, library=None, subtype=None, **parameters):
        """Create cell from an abstract class using the specified sub-class type.

        This is to be called from the ``create()`` function of abstract classes. It takes care of
        creating a code generated or a file based cell.

        Args:
            layout: pya.Layout object where this cell is created
            library: LIBRARY_NAME of the calling PCell instance
            subtype: name (str) of the desired sub-class of ``cls``
            **parameters: PCell parameters for the element as keyword arguments

        Return:
            tuple of the cell instance and a boolean indicating code generated cell
        """

        library_layout = (load_libraries(path=cls.LIBRARY_PATH)[cls.LIBRARY_NAME]).layout()

        if subtype is None:  # derive type from the class name
            subtype = to_library_name(cls.__name__)

        cl = cls._get_abstract()
        module = to_module_name(cl.__name__)
        pname = f"{module}_parameters"
        if pname in parameters:
            jp = dict(json.loads(parameters[pname]).items())
            del parameters[pname], parameters[f"_{pname}"]
            parameters = {**jp, **parameters}

        if subtype in library_layout.pcell_names():   # code generated
            pcell_class = type(library_layout.pcell_declaration(subtype))
            return Element._create_cell(pcell_class, layout, library, **parameters), True
        elif library_layout.cell(subtype):    # manually designed
            return layout.create_cell(subtype, cl.LIBRARY_NAME), False
        else:   # fallback is the default
            return cl.create_subtype(layout, library, cl.default_type, **parameters)

    @classmethod
    def create_with_refpoints(cls, layout, library=None, refpoint_transform=pya.DTrans(), rec_levels=None,
                              **parameters):
        """Convenience function to create cell and return refpoints too.

        Args:
            layout: pya.Layout object where this cell is created
            library: LIBRARY_NAME of the calling PCell instance
            refpoint_transform: transform for converting refpoints into target coordinate system
            rec_levels: recursion level when looking for refpoints from subcells. Set to 0 to disable recursion.
            **parameters: PCell parameters for the element, as keyword argument
        """
        cell = cls.create(layout, library, **parameters)
        refp = get_refpoints(layout.layer(default_layers["refpoints"]), cell, refpoint_transform, rec_levels)
        return cell, refp

    def add_element(self, cls, **parameters):
        """Create a new cell for the given element in this layout.

        Args:
            cls: Element subclass to be created
            **parameters: PCell parameters for the element as keyword arguments

        Returns:
           the created cell
        """
        parameters = self.pcell_params_by_name(cls, **parameters)
        return cls.create(self.layout, library=self.LIBRARY_NAME, **parameters)

    def insert_cell(self, cell, trans=None, inst_name=None, label_trans=None, align_to=None, align=None, rec_levels=0,
                    **parameters):
        """Inserts a subcell into the present cell.

        It will use the given `cell` object or if `cell` is an Element class' name then directly
        take the provided keyword arguments to first create the cell object.

        If `inst_name` given, the refpoints of the cell are added to the `self.refpoints` with `inst_name` as a prefix,
        and also adds a label `inst_name` to "`"labels layer" at the `base` refpoint and `label_trans` transformation.

        Arguments:
            cell: cell object or Element class name
            trans: used transformation for placement. None by default, which places the subcell into the coordinate
                origin of the parent cell. If `align` and `align_to` arguments are used, `trans` is applied to the
                `cell` before alignment transform which allows for example rotation of the `cell` before placement.
            inst_name: possible instance name inserted into subcell properties under `id`. Default is None
            label_trans: relative transformation for the instance name label
            align_to: location in parent cell coordinates for alignment of cell. Can be either string indicating
                the parent refpoint name, `DPoint` or `DVector`. Default is None
            align: name of the `cell` refpoint aligned to argument `align_to`. Default is None
            rec_levels: recursion level when looking for refpoints from subcells. Set to 0 to disable recursion.
            **parameters: PCell parameters for the element, as keyword argument

        Return:
            tuple of placed cell instance and reference points with the same transformation
        """
        if isclass(cell):
            parameters = self.pcell_params_by_name(cell, **parameters)
            cell = cell.create(self.layout, library=self.LIBRARY_NAME, **parameters)

        if isinstance(align_to, str):
            align_to = self.refpoints[align_to]

        cell_inst, refpoints_abs = insert_cell_into(self.cell, cell, trans, inst_name, label_trans, align_to, align,
                                                    rec_levels, **parameters)

        if inst_name is not None:
            # copies probing refpoints to chip level with unique names using subcell id property
            for ref_name, pos in refpoints_abs.items():
                new_name = "{}_{}".format(inst_name, ref_name)
                self.refpoints[new_name] = pos
        return cell_inst, refpoints_abs

    def face(self, face_index=0):
        """Returns the face dictionary corresponding to self.face_ids[face_index].

        The face dictionary contains key "id" for the face ID and keys for all the available layers in that face.

        Args:
            face_index: index of the face_id in self.face_ids, default=0

        """
        return default_faces[self.face_ids[face_index]]

    def pcell_params_by_name(self, cls=None, **parameters):
        """Give PCell parameters as a dictionary.

        Arguments:
            cls: Return only parameters present in this class. All by default.
            **parameters: Optionally update with other keyword arguments

        Returns:
            A dictionary of all PCell parameter names and corresponding current values.
        """
        keys = type(self).get_schema().keys()

        if cls is not None:  # filter keys by cls
            if Element.build == cls.build:  # Abstract class? Find subclass specified by *_type.
                cls = cls._get_abstract()
                mod_type = f"{to_module_name(cls.__name__)}_type"
                subtype = parameters[mod_type] if mod_type in parameters else getattr(self, mod_type, "")
                if subtype:
                    library_layout = (load_libraries(path=cls.LIBRARY_PATH)[cls.LIBRARY_NAME]).layout()
                    if subtype in library_layout.pcell_names():
                        cls = type(library_layout.pcell_declaration(subtype))
            keys = list(set(cls.get_schema().keys()) & set(keys))

        p = {k: self.__getattribute__(k) for k in keys if k != "refpoints"}
        return {**p, **parameters}

    def add_port(self, name, pos, direction=None, face_id=0):
        """ Add a port location to the list of reference points as well as ports layer for netlist extraction

        Args:
            name: name for the port. Will be "decorated" for annotation layer, left as is for port layer. If evaluates
                to False, it will be replaced with `port`
            pos: pya.DVector or pya.DPoint marking the position of the port in the Element base
            direction: direction of the signal going _to_ the port to determine the location of the "corner" reference
                point which is used for waveguide direction. If evaluates to False as is the default, no corner point is
                added.
            face_id: index of the face id, default=0
        """
        text = pya.DText(name, pos.x, pos.y)
        self.cell.shapes(self.get_layer("ports", face_id)).insert(text)

        port_name = name if "port" in name else ("port_"+name if name else "port")
        self.refpoints[port_name] = pos
        if direction:
            self.refpoints[port_name+"_corner"] = pos+direction/direction.length()*self.r

    def copy_port(self, name, cell_inst, new_name=None):
        """ Copy a port definition from a different cell and instance; typically used to expose a specific subcell port.

        Args:
            name: Name of the port as it was specified to ``add_port``
            cell_inst: Instance of the cell, used to transform the port location correctly.
            new_name: Optionally rename the port
        """
        copy_name = name if new_name is None else new_name
        port_name = "port" if name == "" else f"port_{name}"
        port_corner_name = f"{port_name}_corner"

        # workaround for getting the cell due to KLayout bug, see
        # https://www.klayout.de/forum/discussion/1191/cell-shapes-cannot-call-non-const-method-on-a-const-reference
        # TODO: replace by `cell = cell_inst.cell` once KLayout bug is fixed (may be fixed in 0.27 but seems untested)
        cell = self.layout.cell(cell_inst.cell_index)

        cell_refpoints = self.get_refpoints(cell, cell_inst.dcplx_trans)
        for i in range(len(self.face_ids)):
            if "ports" in self.face(i):
                if name in get_refpoints(self.get_layer("ports", i), cell, cell_inst.dcplx_trans):
                    if port_corner_name in cell_refpoints:
                        self.add_port(copy_name, cell_refpoints[port_name],
                                      cell_refpoints[port_corner_name] - cell_refpoints[port_name], i)
                    else:
                        self.add_port(copy_name, cell_refpoints[port_name], face_id=i)
                    break

    @classmethod
    def get_schema(cls, noparents=False, abstract_class=None):
        """Returns the combined parameters of the class "cls" and all its ancestor classes.

        Args:
            noparents: If True then only return the parameters of "cls", not including ancestors.
            abstract_class: Return parameters up to this abstract class if specified.
        """
        schema = {}
        for pc in cls.__mro__:
            if not hasattr(pc, 'LIBRARY_NAME'):
                break
            schema = {**Param.get_all(pc), **schema}
            if noparents or abstract_class == pc:  # not interested in more parent classes
                break
        return schema

    def produce_impl(self):
        """This method builds the PCell.

        Adds all refpoints to user properties and draws their names to the annotation layer.
        """
        self.refpoints = {}

        # Put general "infrastructure actions" here, before build()
        self.refpoints["base"] = pya.DPoint(0, 0)

        self.build()

        self.post_build()

        for name, refpoint in self.refpoints.items():
            text = pya.DText(name, refpoint.x, refpoint.y)
            self.cell.shapes(self.get_layer("refpoints")).insert(text)

    def build(self):
        """Child classes re-define this method to build the PCell."""

    def post_build(self):
        """Child classes re-define this method for post-build operations"""

    def display_text_impl(self):
        if self.display_name:
            return self.display_name
        return type(self).__name__

    def get_refpoints(self, cell, cell_transf=pya.DTrans(), rec_levels=None):
        """See `get_refpoints`."""
        return get_refpoints(self.layout.layer(default_layers["refpoints"]), cell, cell_transf, rec_levels)

    def get_layer(self, layer_name, face_id=0):
        """Returns the specified Layer object.

        Args:
            layer_name: layer name text
            face_id: Name or index of the face to use, default=0
        """
        if isinstance(face_id, str):
            return self.layout.layer(self.face(self.face_ids.index(face_id))[layer_name])
        elif (face_id == 0) and (layer_name not in self.face(0)):
            return self.layout.layer(default_layers[layer_name])
        else:
            return self.layout.layer(self.face(face_id)[layer_name])

    @staticmethod
    def _create_cell(elem_cls, layout, library=None, **parameters) -> pya.Cell:
        """Create cell for elem_cls in layout.

        This is separated from the class method `create` to enable invocation from classes where `create` is shadowed.

        Args:
            elem_cls: element class for which the cell is created
            layout: pya.Layout object where this cell is created
            library: LIBRARY_NAME of the calling PCell instance
            **parameters: PCell parameters for the element as keyword arguments
        """
        cell_library_name = to_library_name(elem_cls.__name__)
        if elem_cls.LIBRARY_NAME == library:  # Matthias' workaround: https://github.com/KLayout/klayout/issues/905
            return layout.create_cell(cell_library_name, parameters)
        else:
            load_libraries(path=elem_cls.LIBRARY_PATH)
            return layout.create_cell(cell_library_name, elem_cls.LIBRARY_NAME, parameters)

    @classmethod
    def _get_abstract(cls):
        """Helper function to return ``cls``'s abstract class, if available, otherwise just return ``cls``."""
        if not hasattr(cls, "default_type"):
            return cls
        prev = cls
        abstract = cls.__bases__[0]
        while hasattr(abstract, "default_type"):
            prev = abstract
            abstract = prev.__bases__[0]
        return prev

    def _add_parameter(self, name, value_type, description,
                       default=None, unit=None, hidden=False, readonly=False, choices=None, docstring=None):
        """Creates a `pya.PCellParameterDeclaration` object and appends it to `self._param_decls`

        The arguments to this function define the PCellParameterDeclaration attributes with the same names,
        except:

            * `value_type` defines the `type` attribute
            * `docstring` is a more verbose parameter description, used in documentation generation.
            * `choices` argument is a list of `(description, value)` tuples. For convenience it also accepts
              self-describing, plain string elements, these will be converted to the expected tuple format.

        For TypeLayer parameters this also defines a `name_layer` read accessor for the layer index and modifies
        `self._layer_param_index` accordingly.
        """
        # pylint: disable=unused-argument

        # special handling of layer parameters
        if value_type == pya.PCellParameterDeclaration.TypeLayer:
            setattr(type(self), name + "_layer",
                    pya._PCellDeclarationHelperLayerDescriptor(len(self._layer_param_index)))
            self._layer_param_index.append(len(self._param_decls))

        # create the PCellParameterDeclaration and add to self._param_decls
        param_decl = pya.PCellParameterDeclaration(name, value_type, description, default, unit)
        param_decl.hidden = hidden
        param_decl.readonly = readonly
        if choices is not None:
            if not isinstance(choices, list) and not isinstance(choices, tuple):
                raise ValueError("choices must be a list or tuple.")
            for choice in choices:
                if  isinstance(choice, str):  # description-is-value shorthand
                    choice = (choice, choice)
                if len(choice) != 2:
                    raise ValueError("Each item in choices list/tuple must be a two-element array [description, value]")
                param_decl.add_choice(choice[0], choice[1])
        self._param_decls.append(param_decl)

    def raise_error_on_cell(self, error_msg, position=pya.DPoint()):
        """Replaces cell with error text in the annotation layer, and raises ValueError with the same error message.

        Args:
             error_msg: the error message
             position: location of the text center (optional)
        """
        self.cell.clear()
        error_text_cell = self.layout.create_cell("TEXT", "Basic", {
            "layer": default_layers["annotations"],
            "text": error_msg,
            "mag": 10.0
        })
        text_center = error_text_cell.bbox().center().to_dtype(self.layout.dbu)
        self.insert_cell(error_text_cell, pya.DTrans(position - text_center))
        raise ValueError(error_msg)

    def add_protection(self, shape, face_id=0):
        """Add ground grid protection shape

        Args:
             shape: The shape (Region, DPolygon, etc.) to add to ground_grid_avoidance layer
             face_id: primary face index of ground_grid_avoidance layer, default=0
        """

        self.cell.shapes(self.get_layer("ground_grid_avoidance", face_id)).insert(shape)
        if self.protect_opposite_face:
            for group in self.opposing_face_id_groups:
                if self.face_ids[face_id] in group:
                    for other_face_id in group:
                        if other_face_id != self.face_ids[face_id] and other_face_id in self.face_ids:
                            self.cell.shapes(self.get_layer("ground_grid_avoidance",
                                                            self.face_ids.index(other_face_id))).insert(shape)
                    break

    def sync_parameters(self, abc):
        """Syncronise the calling class' parameters with a JSON representation.

        This is called several times from coerce_parameters_impl() while using the PCell editor GUI. Particularly, each
        time a parameter of abc's sub-class is changed by the user. It figures out which parameter is changed and
        updates the ``*_parameters`` JSON strings accordingly, or the other way around.

        For example, if abc is Fluxline and the fluxline_width parameter is changed in GUI then the fluxline_parameters
        JSON string will be updated with this value. Or if fluxline_parameters string is changed then the corresponding
        fluxline parameter of the calling pcell is updated.

        Args:
            abc: An abstract class. Only consider parameters of this class' descendants
        """
        if is_standalone_session():
            return

        def pformat(a):
            if a.data_type == pdt.TypeInt:
                return int(a.default)
            elif a.data_type == pdt.TypeDouble:
                return float(a.default)
            return a.default

        module = to_module_name(abc.__name__)
        pname = f"{module}_parameters"
        json_str = getattr(self, pname)
        saved = getattr(self, f"_{pname}")
        params = json.loads(json_str) if json_str else {}
        subtype = getattr(self, f"{module}_type")
        pd = {}

        if subtype == "none":
            return
        if  json_str == "{}" or params[f"{module}_type"] != subtype:  # initialise defaults
            if f"{module}_type" in params and json_str != saved and saved != {}:
                subtype = params[f"{module}_type"]
                setattr(self, f"{module}_type", subtype)
            cls = element_by_class_name(subtype.replace(" ", ""), abc.LIBRARY_PATH, abc.LIBRARY_NAME)
            schema = cls.get_schema(abstract_class=abc)
            for k, v in schema.items():
                if not k.endswith("_parameters"):
                    pd[k] = getattr(self, k) if hasattr(self, k) else pformat(v)
            json_str = json.dumps(pd)
            setattr(self, pname, json_str)
            Param.get_all(abc)[pname].default = json_str
        elif saved == json_str:  # some parameters changed, update the JSON string
            for k, v in params.items():
                pd[k] = getattr(self, k) if hasattr(self, k) else v
            json_str = json.dumps(pd)
            setattr(self, pname, json_str)
        else:  # the JSON string changed, use it to update other parameters
            for k, v in params.items():
                if hasattr(self, k):
                    setattr(self, k, v)
        setattr(self, f"_{pname}", json_str)

    @classmethod
    def get_sim_ports(cls, simulation): # pylint: disable=unused-argument
        """List of RefpointToSimPort objects defining which refpoints
        should be turned to simulation ports for the given element class

        Returns empty list if not implemented for Element subclass.
        When implementing this method, the best practice is for this method
        to have no "side effects", that is all code contained within this method
        should only serve to derive the list of RefpointToSimPort objects and nothing
        else: no change in element's geometry or parameter values.

        Args:
            cls: Element class, this is a class method
            simulation: Simulation object where a cell of this element class is placed.
                Use this argument if you need to decide certain arguments
                for RefpointToSimPort objects based on simulation's parameters

        Returns:
            List of RefpointToSimPort objects, empty list by default
        """
        return []

    @staticmethod
    def left_and_right_waveguides(simulation):
        """A common implementation of get_sim_ports that adds left and right waveguides
        to port_a and port_b respectively. The a and b values of right waveguide
        can be adjusted separately.
        """
        a2 = simulation.a if simulation.a2 < 0 else simulation.a2
        b2 = simulation.b if simulation.b2 < 0 else simulation.b2
        return [WaveguideToSimPort("port_a", side='left', a=simulation.a, b=simulation.b),
                WaveguideToSimPort("port_b", side='right', a=a2, b=b2)]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from math import ceil

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.spiral_resonator_polygon import SpiralResonatorPolygon
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter
from kqcircuits.pya_resolver import pya
from kqcircuits.util.coupler_lib import cap_params
from kqcircuits.util.geometry_helper import get_angle
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.chips.quality_factor import QualityFactor


@add_parameters_from(QualityFactor, "res_lengths", "n_fingers", "l_fingers", "type_coupler",
                     res_a=[10, 10, 10, 20, 10, 5], res_b=[6, 6, 6, 12, 6, 3])
@add_parameters_from(Chip, frames_enabled=[0, 1])
@add_parameters_from(SpiralResonatorPolygon, "bridge_spacing")
class QualityFactorTwoface(Chip):
    """The PCell declaration for a QualityFactorTwoFace chip.

    Preliminary class for flip-chip resonators.
    """
    resonator_types = Param(pdt.TypeList, "Resonator types (capped, twoface, etched, or solid)", ["capped"] * 6,
                            docstring="Choices: 'capped', 'twoface', 'etched', 'solid'")
    resonator_faces = Param(pdt.TypeList, "Resonator face order list", [0, 1])
    connector_distances = Param(pdt.TypeList, "Resonator input to face to face connector",
                                [500, 1300, 2100, 2900, 3700, 4500] * 2, unit="[μm]",
                                docstring="Distances of face to face connectors from resonator inputs")
    spiral_box_height = Param(pdt.TypeDouble, "Spiral resonator box height", 2000)
    spiral_box_width = Param(pdt.TypeDouble, "Spiral resonator box width", 500)
    x_indentation = Param(pdt.TypeDouble, "Resonator/connector indentation from side edges", 800)
    cap_res_distance = Param(pdt.TypeDouble, "Distance between spiral resonator and capacitor", 200)
    waveguide_indentation = Param(pdt.TypeDouble, "Waveguide indentation from top chip edge", 500)
    extra_resonator_avoidance = Param(pdt.TypeList, "Added avoidance", [0, 0, 0, 0, 0, 0], unit="[μm]",
                                      docstring="Added avoidance around resonators [μm]")
    etch_opposite_face_margin = Param(pdt.TypeDouble, "Margin around the waveguide to etch on the opposite face " +
                                                      "for 'etched' type resonators", 5)

    def build(self):
        self._produce_resonators()

    def _produce_resonators(self):
        self.produce_launchers("SMA8", launcher_assignments={
            3: "PL-1-OUT",
            8: "PL-1-IN",
        })

        # Constants
        face1_box = self.get_box(1)
        left_x = face1_box.left + self.waveguide_indentation
        right_x = face1_box.right - self.waveguide_indentation
        left_connector = face1_box.left + self.x_indentation
        right_connector = face1_box.right - self.x_indentation
        mid_y = face1_box.center().y
        resonator_face_ids = [self.face_ids[int(i)] for i in self.resonator_faces]

        n_resonators = len(self.res_lengths)
        start_x = left_connector + self.spiral_box_width
        resonator_spacing = (right_connector - left_connector - self.spiral_box_width) / n_resonators
        resonator_positions = [pya.DPoint(start_x + (i + 0.5) * resonator_spacing, mid_y) for i in range(n_resonators)]
        resonator_directions = [90, 270] * ceil(n_resonators / 2)

        # Create probeline
        tee_length = self.a_capped / 2 + self.b_capped
        side_tee_length = tee_length + 5 * self.a_capped
        splitter_nodes = [Node(point, WaveguideCoplanarSplitter, angles=[0, 180, angle],
                               lengths=[tee_length, tee_length, side_tee_length], inst_name=f"{i}_t")
                          for i, (point, angle) in enumerate(zip(resonator_positions, resonator_directions))]

        probeline = self.add_element(
            WaveguideComposite,
            nodes=[
                Node(self.refpoints["PL-1-IN_port"]),
                Node(pya.DPoint(face1_box.left, self.refpoints["PL-1-IN_port"].y), a=self.a_capped, b=self.b_capped,
                     taper_length=100),
                Node(pya.DPoint(left_x, self.refpoints["PL-1-IN_port"].y)),
                Node(pya.DPoint(left_x, mid_y)),
                Node(pya.DPoint(left_connector, mid_y), face_id=resonator_face_ids[0]),
                *splitter_nodes,
                Node(pya.DPoint(right_connector, mid_y), face_id=self.face_ids[0]),
                Node(pya.DPoint(right_x, mid_y)),
                Node(pya.DPoint(right_x, self.refpoints["PL-1-OUT_port"].y)),
                Node(pya.DPoint(face1_box.right - 100, self.refpoints["PL-1-OUT_port"].y), a=self.a, b=self.b,
                     taper_length=100),
                Node(self.refpoints["PL-1-OUT_port"])
            ],
        )
        self.insert_cell(probeline, inst_name="pl")

        # Create resonators
        for i in range(n_resonators):
            self.produce_resonator(i, float(self.res_a[i]), float(self.res_b[i]), float(self.res_lengths[i]),
                                   float(self.n_fingers[i]), float(self.l_fingers[i]), self.type_coupler[i],
                                   resonator_face_ids, self.resonator_types[i], float(self.connector_distances[i]),
                                   float(self.extra_resonator_avoidance[i]), mirror=(i % 2 == 1))

    def produce_resonator(self, i, a, b, length, n_fingers, l_fingers, type_coupler, face_ids,
                          resonator_type="capped", connector_distance=0.0,
                          extra_resonator_avoidance=0.0, mirror=False):
        """
        Produce a single spiral resonator and corresponding input capacitor.

        The resonator is attached to the chip refpoint ``pl_{i}_t_port_c``, where `i` is the resonator index (this can
        be any identifier, as long as it is used the same way in the refpoint name). The direction of the resonator
        is set by the corresponding ``_corner`` refpoint, but the spiral can be left or right depending on ``mirror``.

        Args:
            i: Resonator index
            a: CPW line width
            b: CPW gap width
            length: Resonator length (excluding capacitor)
            n_fingers: Capacitor finger number, or finger_control for ``SmoothCapacitor``
            l_fingers: Capacitor finger length
            type_coupler: Type of capacitor, see ``cap_params``
            face_ids: list of face ids for the resonator
            resonator_type: String, type of resonator, one of ``etched``, ``capped``, ``solid`` or ``twoface``
            connector_distance: For ``twoface`` resonators, distance of the flip chip connector starting from capacitor
            extra_resonator_avoidance: Extra ``ground_grid_avoidance`` margin around the resonator
            mirror: Turn clockwise if False, or counter-clockwise if True.
        """

        if resonator_type in ["etched", "capped"]:
            protect_opposite_face = False
        elif resonator_type == "solid":
            protect_opposite_face = True
        else:
            protect_opposite_face = self.protect_opposite_face

        # Get starting angle
        start = self.refpoints[f"pl_{i}_t_port_c"]
        start_corner = self.refpoints[f"pl_{i}_t_port_c_corner"]
        start_angle = get_angle(start_corner - start)

        # Capacitor
        cplr_params = cap_params(
            n_fingers, l_fingers, type_coupler, protect_opposite_face=protect_opposite_face,
            face_ids=face_ids, a=self.a_capped, b=self.b_capped, a2=a, b2=b, element_key='cell')
        inst_cplr, cplr_refpoints = self.insert_cell(**cplr_params, align="port_a", align_to=start,
                                                     trans=pya.DCplxTrans(1, start_angle, False, 0, 0))

        # Spiral resonator
        if resonator_type == "twoface":
            res_params = {'connector_dist': connector_distance, "bridge_spacing": 0}
        elif resonator_type == "etched":
            res_params = {
                "airbridge_type": "Airbridge Multi Face",
                "include_bumps": False,
                "bridge_length": a + 2 * (b + self.etch_opposite_face_margin + extra_resonator_avoidance),
                "bridge_width": 2,
                "pad_length": 2,
                "bridge_spacing": self.bridge_spacing,
            }
        else:
            res_params = {"bridge_spacing": 0}

        inst_res, _ = self.insert_cell(
            SpiralResonatorPolygon,
            margin=self.margin + extra_resonator_avoidance,
            input_path=pya.DPath([
                pya.DPoint(0, 0),
            ], 10),
            poly_path=pya.DPath([
                pya.DPoint(self.cap_res_distance, 0),
                pya.DPoint(self.spiral_box_height, 0),
                pya.DPoint(self.spiral_box_height, -self.spiral_box_width),
                pya.DPoint(self.cap_res_distance, -self.spiral_box_width)
            ], 10),
            length=length,
            a=a,
            b=b,
            face_ids=face_ids,
            protect_opposite_face=protect_opposite_face,
            **res_params,
            inst_name=f'resonator{i}',
            trans=pya.DCplxTrans(1, start_angle, mirror, cplr_refpoints["port_b"])
        )

        # Top chip etching and grid avoidance above resonator
        if resonator_type == "etched":
            l0 = self.get_layer("ground_grid_avoidance", int(self.resonator_faces[0]))
            region = pya.Region(inst_res.cell.begin_shapes_rec(l0)).transformed(inst_res.trans)
            region += pya.Region(inst_cplr.cell.begin_shapes_rec(l0)).transformed(inst_cplr.trans)
            region = region.sized((self.etch_opposite_face_margin - self.margin) / self.layout.dbu)
            protection_region = region.sized(self.margin / self.layout.dbu)
            opposite_face = int(self.resonator_faces[1])
            self.cell.shapes(self.get_layer("ground_grid_avoidance", opposite_face)).insert(protection_region)
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid", opposite_face)).insert(region)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.chips.lithography_test import LithographyTest
from kqcircuits.chips.chip import Chip
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import add_parameters_from

@add_parameters_from(Chip, frames_enabled=[0, 1])
@add_parameters_from(LithographyTest, 'stripe_test')
class LithographyTestTwoface(Chip):
    """Optical lithography test chip in a flip chip architecture.

    Consists of StripesTest cells with different parameters.
    """
    create_pattern = LithographyTest.create_pattern

    def build(self):
        cell_horizontal_1, cell_vertical_1, cell_diagonal_1 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=1,
                                                                                  face_id=self.face_ids[0])
        cell_horizontal_2, cell_vertical_2, cell_diagonal_2 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=5,
                                                                                  face_id=self.face_ids[0])
        cell_horizontal_3, cell_vertical_3, cell_diagonal_3 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=1,
                                                                                  face_id=self.face_ids[1])
        cell_horizontal_4, cell_vertical_4, cell_diagonal_4 = self.create_pattern(num_stripes=20, length=100,
                                                                                  min_width=1,
                                                                                  max_width=15, step=1, spacing=5,
                                                                                  face_id=self.face_ids[1])
        # bottom patterns
        self.insert_cell(cell_horizontal_1, pya.DCplxTrans(1, 0, False, 2000, 6500))
        self.insert_cell(cell_horizontal_1, pya.DCplxTrans(1, 0, False, 3000, 6500))
        self.insert_cell(cell_horizontal_2, pya.DCplxTrans(1, 0, False, 4000, 6500))
        self.insert_cell(cell_horizontal_2, pya.DCplxTrans(1, 0, False, 6000, 6500))
        self.insert_cell(cell_vertical_1, pya.DCplxTrans(1, 0, False, 1500, 5700))
        self.insert_cell(cell_vertical_1, pya.DCplxTrans(1, 0, False, 5500, 5700))
        self.insert_cell(cell_vertical_2, pya.DCplxTrans(1, 0, False, 1500, 3800))
        self.insert_cell(cell_vertical_2, pya.DCplxTrans(1, 0, False, 5500, 3800))
        self.insert_cell(cell_diagonal_1, pya.DCplxTrans(1, 0, False, 400, 1500))
        self.insert_cell(cell_diagonal_1, pya.DCplxTrans(1, 0, False, 1500, 1500))
        self.insert_cell(cell_diagonal_2, pya.DCplxTrans(1, 0, False, 3100, 1500))
        self.insert_cell(cell_diagonal_2, pya.DCplxTrans(1, 0, False, 6100, 1500))
        # top patterns
        self.insert_cell(cell_horizontal_3, pya.DCplxTrans(1, 0, False, 5900, 4000) * pya.DCplxTrans.M90)
        self.insert_cell(cell_horizontal_4, pya.DCplxTrans(1, 0, False, 8000, 4000) * pya.DCplxTrans.M90)
        self.insert_cell(cell_vertical_3, pya.DCplxTrans(1, 0, False, 5000, 6200) * pya.DCplxTrans.M90)
        self.insert_cell(cell_vertical_4, pya.DCplxTrans(1, 0, False, 5000, 4000) * pya.DCplxTrans.M90)
        self.insert_cell(cell_diagonal_3, pya.DCplxTrans(1, 0, False, 7500, 1800) * pya.DCplxTrans.M90)
        self.insert_cell(cell_diagonal_4, pya.DCplxTrans(1, 0, False, 6200, 1800) * pya.DCplxTrans.M90)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.geometry_helper import get_angle
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.defaults import default_airbridge_type, default_layers
from kqcircuits.elements.airbridges import airbridge_type_choices
from kqcircuits.elements.element import Element, get_refpoints


class Airbridge(Element):
    """Airbridge base class without actual produce function.

    All subclasses should have ports 'a' and 'b' at positions (0, bridge_length) and (0, -bridge_length).
    """

    default_type = default_airbridge_type
    """This is the default shape if not specified otherwise by the user."""

    airbridge_type = Param(pdt.TypeString, "Airbridge type", default_type, choices=airbridge_type_choices)
    bridge_width = Param(pdt.TypeDouble, "Bridge width", 20, unit="μm")
    pad_length = Param(pdt.TypeDouble, "Pad length", 18, unit="μm")
    bridge_length = Param(pdt.TypeDouble, "Bridge length (from pad to pad)", 44, unit="μm")

    @classmethod
    def create(cls, layout, library=None, airbridge_type=None, **parameters):
        """Create cell for an airbridge in layout."""
        cell, code_generated = cls.create_subtype(layout, library, airbridge_type, **parameters)

        # transform cell to have 'port_a' at (0, l/2) and 'port_b' at (0, -l/2), where l is distance between ports
        if not code_generated:
            ref_points = get_refpoints(layout.layer(default_layers["refpoints"]), cell)
            center = (ref_points['port_a'] + ref_points['port_b']) / 2
            orientation = get_angle(ref_points['port_a'] - ref_points['port_b'])
            cell.transform(pya.DCplxTrans(1.0, 90 - orientation, False, -center))

        return cell

    def _produce_bottom_pads(self, pts):
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("airbridge_pads")).insert(shape)

        # bottom layer lower pad
        self.cell.shapes(self.get_layer("airbridge_pads")).insert(pya.DTrans.M0 * shape)

        # protection layer
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(shape.sized(self.margin))
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(pya.DTrans.M0 * shape.sized(self.margin))

        # refpoints for connecting to waveguides
        self.add_port("a", pya.DPoint(0, self.bridge_length / 2), pya.DVector(0, 1))
        self.add_port("b", pya.DPoint(0, -self.bridge_length / 2), pya.DVector(0, -1))

    def _produce_top_pads_and_bridge(self, pts):
        shape = pya.DPolygon(pts)
        self.cell.shapes(self.get_layer("airbridge_flyover")).insert(shape)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.test_structures.test_structure import TestStructure


class StripesTest(TestStructure):
    """PCell declaration for optical lithography test stripes.

    Contains a given number of lines in `base_metal_gap_wo_grid`-layer with given width, length and spacing.
    There is also a text in `base_metal_gap_wo_grid`-layer next to the lines, which shows the line width.
    """

    num_stripes = Param(pdt.TypeInt, "Number of stripes", 20)
    stripe_width = Param(pdt.TypeDouble, "Width of the stripes", 1, unit="μm")
    stripe_length = Param(pdt.TypeDouble, "Length of the stripes", 100, unit="μm")
    stripe_spacing = Param(pdt.TypeDouble, "Spacing between the stripes", 1, unit="μm")

    def build(self):

        layer_base_metal = self.get_layer("base_metal_gap_wo_grid")

        width = float(self.stripe_width)
        stripe = pya.DBox(0, 0, width, self.stripe_length)

        for i in range(self.num_stripes):
            trans = pya.DTrans(i*(width + self.stripe_spacing), 0)
            self.cell.shapes(layer_base_metal).insert(trans * stripe)

        width_str = int(width) if width.is_integer() else width
        text_cell = self.layout.create_cell("TEXT", "Basic", {
            "layer": self.face()["base_metal_gap_wo_grid"],
            "text": "{}".format(width_str),
            "mag": 40,
        })
        text_x = self.num_stripes*(width + self.stripe_spacing) + width
        text_y = self.stripe_length/2
        self.insert_cell(text_cell, pya.DTrans(text_x, text_y))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.util.parameters import Param, pdt
from kqcircuits.qubits.qubit import Qubit
from kqcircuits.pya_resolver import pya
from kqcircuits.util.refpoints import JunctionSimPort


class Swissmon(Qubit):
    """The PCell declaration for a Swissmon qubit.

    Swissmon type qubit. Each arm (West, North, East, South) has it's own arm gap width
    (``gap_width``) and arm metal width (``arm_width``). SQUID is loaded from another library.
    Option of having fluxline.  Refpoints for 3 couplers, fluxline position and chargeline position.
    Length between the ports is from waveguide port to the rectangular part of the launcher pad.
    Length of the fingers is also used for the length of the launcher pad.

    .. MARKERS_FOR_PNG 56,-61 140,0 0,175 -64,117
    """

    arm_length = Param(pdt.TypeList, "Arm length (um, WNES))", [300. / 2] * 4)
    arm_width = Param(pdt.TypeList, "Arm metal width (um, WNES)", [24, 24, 24, 24])
    gap_width = Param(pdt.TypeList, "Arm gap width (um, WNES)", [12, 12, 12, 12])
    cpl_width = Param(pdt.TypeList, "Coupler width (um, WNE)", [24, 24, 24])
    cpl_length = Param(pdt.TypeList, "Coupler lengths (um, WNE)", [120, 120, 120])
    cpl_gap = Param(pdt.TypeList, "Coupler gap (um, WNE)", [102, 102, 102])
    port_width = Param(pdt.TypeList, "Port width (um, WNE)", [10, 10, 10])
    cl_offset = Param(pdt.TypeList, "Chargeline offset (um, um)", [200, 200])
    island_r = Param(pdt.TypeDouble, "Center island rounding radius", 5, unit="μm")

    def build(self):
        self._produce_cross_and_squid()

        self._produce_chargeline()  # refpoint only ATM

        self.produce_fluxline()

        for i in range(3):
            self._produce_coupler(i)

    def _produce_chargeline(self):
        # shorthands
        l = [float(offset) for offset in self.cl_offset]  # swissmon arm length from the center of the cross (refpoint)

        # add ref point
        # port_ref = pya.DPoint(-g-b-a/2, -l)
        port_ref = pya.DPoint(-l[0], -l[1])
        self.add_port("drive", port_ref)

    def _produce_coupler(self, cpl_nr):
        # shorthand
        a = float(self.port_width[cpl_nr])
        b = self.b
        [ww, wn, we, ws] = [float(width) / 2 for width in self.arm_width]
        aw = [ww, wn, we, ws][cpl_nr]
        w = float(self.cpl_width[cpl_nr])
        l = float(self.cpl_length[cpl_nr])
        g = float(self.cpl_gap[cpl_nr]) / 2

        # Location for connecting the waveguides to
        port_shape = pya.DBox(-a / 2, 0, a / 2, b)
        port_region = pya.Region([port_shape.to_itype(self.layout.dbu)])

        if l > 0:
            # Horseshoe opened to below
            # Refpoint in the top center
            shoe_points = [
                pya.DPoint(a, 0),
                pya.DPoint(g + w, 0),
                pya.DPoint(g + w, -l),
                pya.DPoint(g, -l),
                pya.DPoint(g, -w),
                pya.DPoint(-g, -w),
                pya.DPoint(-g, -l),
                pya.DPoint(-g - w, -l),
                pya.DPoint(-g - w, 0),
                pya.DPoint(-a, 0)
            ]
            shoe = pya.DPolygon(shoe_points)
            shoe.size(b)
            shoe.insert_hole(shoe_points[::-1])

            # convert to range and recover CPW port
            shoe_region = pya.Region([shoe.to_itype(self.layout.dbu)])
            shoe_region.round_corners(self.island_r / self.layout.dbu, self.island_r / self.layout.dbu, self.n)
            shoe_region2 = shoe_region - port_region

        # move to the north arm of swiss cross
        ground_width = (2 * g - float(self.gap_width[1]) - 2 * b) / 2
        shift_up = float(self.arm_length[cpl_nr]) + (float(self.gap_width[1]) - 2 * aw) / 2 + ground_width + w + b
        transf = pya.DCplxTrans(1, 0, False, pya.DVector(0, shift_up))

        # rotate to the correct direction
        rotation = [
            pya.DCplxTrans.R90, pya.DCplxTrans.R0, pya.DCplxTrans.R270
        ][cpl_nr]

        # draw
        if l > 0:
            self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(
                shoe_region2.transformed((rotation * transf).to_itrans(self.layout.dbu)))
        self.cell.shapes(self.get_layer("waveguide_path")).insert(
            port_region.transformed((rotation * transf).to_itrans(self.layout.dbu)))

        # protection
        if l > 0:
            protection = pya.DBox(-g - w - b - self.margin, -l - b - self.margin, g + w + b + self.margin,
                                  b + self.margin)
            self.cell.shapes(self.get_layer("ground_grid_avoidance")).\
                insert(protection.transformed((rotation * transf)))

        # add ref point
        port_ref = pya.DPoint(0, b)
        self.add_port("cplr{}".format(cpl_nr), (rotation * transf).trans(port_ref), rotation*pya.DVector(0, 1))

    def _produce_cross_and_squid(self):
        """Produces the cross and squid for the Swissmon."""
        # shorthand
        [ww, wn, we, ws] = [float(width) / 2 for width in self.arm_width]
        l = [float(length) for length in self.arm_length]

        [sw, sn, se, ss] = [float(width) for width in self.gap_width]

        # # SQUID
        # SQUID origin at the ground plane edge
        squid_transf = pya.DCplxTrans(1, 0, False, pya.DVector(0, -l[3] - ss))
        squid_ref_rel = self.produce_squid(squid_transf)
        # SQUID port_common at the end of the south arm
        squid_length = squid_ref_rel["port_common"].distance(pya.DPoint(0, 0))

        # Swissmon etch region

        # refpoint in the center of the swiss cross
        cross_island_points = [
            pya.DPoint(wn, we),
            pya.DPoint(l[2], we),
            pya.DPoint(l[2], -we),
            pya.DPoint(ws, -we),
            pya.DPoint(ws, -l[3] - ss + squid_length),
            pya.DPoint(-ws, -l[3] - ss + squid_length),
            pya.DPoint(-ws, -ww),
            pya.DPoint(-l[0], -ww),
            pya.DPoint(-l[0], ww),
            pya.DPoint(-wn, ww),
            pya.DPoint(-wn, l[1]),
            pya.DPoint(wn, l[1]),
        ]

        # refpoint in the center of the swiss cross
        cross_gap_points = [
            pya.DPoint(wn + sn, we + se),
            pya.DPoint(l[2] + se, we + se),
            pya.DPoint(l[2] + se, -we - se),
            pya.DPoint(ws + ss, -we - se),
            pya.DPoint(ws + ss, -l[3] - ss),
            pya.DPoint(-ws - ss, -l[3] - ss),
            pya.DPoint(-ws - ss, -ww - sw),
            pya.DPoint(-l[0] - sw, -ww - sw),
            pya.DPoint(-l[0] - sw, ww + sw),
            pya.DPoint(-wn - sn, ww + sw),
            pya.DPoint(-wn - sn, l[1] + sn),
            pya.DPoint(wn + sn, l[1] + sn),
        ]

        cross = pya.DPolygon(cross_gap_points)
        cross.insert_hole(cross_island_points)
        cross_rounded = cross.round_corners(self.island_r, self.island_r, self.n)
        region_etch = pya.Region([cross_rounded.to_itype(self.layout.dbu)])
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(region_etch)

        # Protection
        cross_protection = pya.DPolygon([
            p + pya.DVector(math.copysign(max([sw, sn, se, ss]) + self.margin, p.x),
            math.copysign(max([sw, sn, se, ss]) + self.margin, p.y)) for p in cross_gap_points
        ])
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(cross_protection)

        # Probepoint
        probepoint = pya.DPoint(0, 0)
        self.refpoints["probe_qb_c"] = probepoint

    @classmethod
    def get_sim_ports(cls, simulation):
        return [JunctionSimPort()]



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import importlib
import sys
from pathlib import Path


def module_from_file(file_path):
    """Loads a Python module from the given path.

    Args:
        file_path: Path to a Python file

    Returns:
        The module corresponding to the specified Python file
    """
    file_name = Path(file_path).name[:-3]
    spec = importlib.util.spec_from_file_location(f"{file_name}", file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[f"{file_name}"] = module
    spec.loader.exec_module(module)
    return module



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import sys

from kqcircuits.util.dependencies import install_kqc_dependencies

# pylint: disable=undefined-variable
from ._version import __version__
if hasattr(sys.modules[__name__], "_version"):
    del _version  # remove to avoid confusion with __version__

# Check for needed dependencies and install if missing.
install_kqc_dependencies()



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import logging
import sys


def route_log(lowest_visible_level="INFO", remove_old_handlers=True, filename="", stdout=True):
    """Routes the log output to stdout and/or the specified logfile.

    Enables monitoring the log in KLayout console. If requested it also appends logs to the
    specified file. By default removes old handlers from root logger to avoid output to other places
    than defined here.

    Arguments:
        lowest_visible_level (String): one of DEBUG, INFO, WARNING, ERROR, CRITICAL
        remove_old_handlers (Boolean): determines if old handlers are removed from the root logger
        filename: name of the file to append logs to
        stdout: enable printing logs to standard output
    """

    root_logger = logging.getLogger()

    # To make sure that the logs are only output to the place defined here, old handlers must be removed from the
    # logger. This is needed for example if you want to set a different lowest_visible_level by re-running this.
    # Otherwise the old handlers would cause output with the old lowest_visible_level.
    if remove_old_handlers:
        while root_logger.hasHandlers():
            root_logger.removeHandler(root_logger.handlers[0])

    message_format = "%(asctime)s:%(levelname)s:%(module)s:%(funcName)s:%(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"
    formatter = logging.Formatter(message_format, date_format)

    if stdout:
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(formatter)
        handler.setLevel(lowest_visible_level)
        root_logger.addHandler(handler)

    if filename:
        fh = logging.FileHandler(filename)
        fh.setFormatter(formatter)
        fh.setLevel(lowest_visible_level)
        root_logger.addHandler(fh)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.defaults import default_fluxline_type
from kqcircuits.elements.fluxlines import fluxline_type_choices


class Fluxline(Element):
    """Base class for fluxline objects without actual produce function."""

    default_type = default_fluxline_type

    fluxline_type = Param(pdt.TypeString, "Fluxline Type", default_type, choices=fluxline_type_choices)
    fluxline_width = Param(pdt.TypeDouble, "Fluxline width", 18, unit="μm")
    fluxline_gap_width = Param(pdt.TypeDouble, "Fluxline gap width", 2, unit="μm")

    fluxline_parameters = Param(pdt.TypeString, "Extra Fluxline Parameters", "{}")
    _fluxline_parameters = Param(pdt.TypeString, "Previous state of *_parameters", "{}", hidden=True)

    @classmethod
    def create(cls, layout, library=None, fluxline_type=None, **parameters):
        """Create a Fluxline cell in layout."""
        return cls.create_subtype(layout, library, fluxline_type, **parameters)[0]

    def coerce_parameters_impl(self):
        self.sync_parameters(Fluxline)

    def _insert_fluxline_shapes(self, left_gap, right_gap):
        """Inserts the gap shapes to the cell.

        Protection layer is added based on their bounding box.

        Arg:
            left_gap (DPolygon): polygon for the left gap
            right_gap (DPolygon): polygon for the right gap
        """
        # transfer to qubit coordinates
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(left_gap)
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(right_gap)
        # protection
        protection = pya.Region([p.to_itype(self.layout.dbu) for p in [right_gap, left_gap]]
                                ).bbox().enlarged(self.margin/self.layout.dbu, self.margin/self.layout.dbu)
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(pya.Polygon(protection))

    def _add_fluxline_refpoints(self, port_ref):
        """Adds refpoints for "port_flux" and "port_flux_corner".

        Arg:
            port_ref (DPoint): position of "port_flux" in fluxline coordinates
        """
        self.refpoints["origin_fluxline"] = pya.DPoint(0, 0)
        self.add_port("flux", port_ref, pya.DVector(0, -1))



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.test_structures.test_structure import TestStructure


class CrossTest(TestStructure):
    """PCell declaration for optical lithography test alignment cross markers.

    Contains a given number of alignment cross markers in `base_metal_gap_wo_grid`-layer with given width, length and
     spacing.
    """

    num_crosses = Param(pdt.TypeInt, "Number of crosses", 10)
    cross_width = Param(pdt.TypeDouble, "Width of the crosses arms", 4, unit="μm")
    cross_length = Param(pdt.TypeDouble, "Length of the crosses", 15, unit="μm")
    cross_spacing = Param(pdt.TypeDouble, "Spacing between the crosses", 100, unit="μm")
    cross_box_distance = Param(pdt.TypeDouble, "Distance between crosses and respective boxes", 4, unit="μm")

    def build(self):

        layer_base_metal = self.get_layer("base_metal_gap_wo_grid")

        box = pya.DBox(pya.DPoint(-self.cross_width/2, -self.cross_box_distance), pya.DPoint(self.cross_width/2,
                                                                    -self.cross_box_distance - self.cross_width))
        vertical_tick = pya.DBox(pya.DPoint(-self.cross_width/2, 0), pya.DPoint(self.cross_width/2, self.cross_length))
        horizontal_tick = pya.DBox(pya.DPoint(-self.cross_length/2, self.cross_length/2 - self.cross_width/2),
                                   pya.DPoint(self.cross_length/2, self.cross_length/2 + self.cross_width/2))

        for i in range(self.num_crosses):
            trans = pya.DTrans(0, False, i*self.cross_spacing, 0).to_itype(self.layout.dbu)
            box_trans = pya.Region(trans * box.to_itype(self.layout.dbu))
            vertical_tick_trans = pya.Region(trans * vertical_tick.to_itype(self.layout.dbu))
            horizontal_tick_trans = pya.Region(trans * horizontal_tick.to_itype(self.layout.dbu))

            marker = box_trans + vertical_tick_trans + horizontal_tick_trans

            self.cell.shapes(layer_base_metal).insert(marker)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.smooth_capacitor import SmoothCapacitor


def cap_params(fingers, length=None, coupler_type="interdigital", element_key='cls', **kwargs):
    """An utility function to easily produce typical finger capacitor instance parameters.
    Covers FingerCapacitorSquare and SmoothCapacitor.

    Args:
        fingers: number of fingers in FingerCapacitorSquare or finger control parameter in SmoothCapacitor.
        length: length of fingers in FingerCapacitorSquare (useless parameter in SmoothCapacitor)
        coupler_type: a string describing the capacitor type (accepts "interdigital", "gap", "ground gap", or "smooth")
        element_key: dictionary key in which coupler Element is returned (use 'cls'=default with `add_element` function
                     or 'cell' with `insert_cell` function)
        **kwargs: other optional parameters

    Returns:
        dictionary of coupler parameters
    """
    if coupler_type == 'smooth':
        return {element_key: SmoothCapacitor,
                'finger_control': fingers,
                "finger_width": 10,
                "ground_gap": 10,
                "finger_gap": 5,
                **kwargs}

    defaults = {element_key: FingerCapacitorSquare,
                "finger_number": int(fingers),
                "finger_length": length,
                "finger_gap_end": 5,
                "finger_gap": 5,
                "finger_width": 15,
                "ground_padding": 10,
               }

    params = {}

    if coupler_type == "gap":
        params = {"finger_length": 0,
                  "finger_gap_end": length,
                  "finger_gap": 0,
                  "finger_width": 10,
                 }
    elif coupler_type == "ground gap":
        params = {"finger_length": 0,
                  "finger_gap_end": length,
                  "finger_gap": 50 - length / 2,
                  "finger_width": 20,
                  "ground_gap_ratio": 1/3
                  }

    return {**defaults, **params, **kwargs}



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from kqcircuits.masks.mask_layout import MaskLayout


class MultiFaceMaskLayout:
    """Class representing multiple mask layouts, corresponding to multiple faces on the same wafer.

    This is a helper class to create multiple ``MaskLayout`` instances, one for each face, and set the same properties
    and chips map for each, and a container for the created mask layouts. It also provides `add_chips_map` that
    distributes over each containing `MaskLayout.add_chips_map`.

    The usual way to instantiate ``MultiFaceMaskLayout`` is through ``MaskSet.add_multi_face_mask_layout``.

    Attributes:
        face_ids: List of face ids to include in this mask layout
        mask_layouts: Dictionary of {face_id: mask_layout} of the individual ``MaskLayouts`` contained in this class
    """
    def __init__(self, layout, name, version, with_grid, face_ids, chips_map=None, extra_face_params=None,
                 mask_layout_type=MaskLayout, **kwargs):
        """Create a multi face mask layout, which can be used to make masks with matching chip maps on multiple faces.

        A ``MaskLayout`` is created of each face in ``face_ids``. If ``face_ids`` is a list, the individual mask layouts
        all have identical parameters. To specify some parameters differently for each mask layout, supply ``face_ids``
        as a dictionary ``{face_ids: extra_params}``, where ``extra_params`` is a dictionary of arguments passed only
        to the mask layout for that face id. These override ``kwargs`` if they contain the same keys.

        By default, ``bbox_face_ids`` is set to ``list(face_ids)`` for all mask layouts.

        Args:
            layout: Layout to use
            name: name of the mask
            version: version of the mask
            with_grid: if True, ground grids are generated
            face_ids: either a list of face ids to include, or a dictionary of ``{face_id: extra_params}``, where
                ``extra_params`` is a dictionary of keyword arguments to apply only to this mask layout.
            chips_map: Chips map to use, or None to use an empty chips map.
            mask_layout_type: optional subclass of MaskLayout to use
            kwargs: any keyword arguments are passed to all containing mask layouts.
        """
        self.face_ids = face_ids
        self.mask_layouts = {}

        for face_id in face_ids:
            all_kwargs = {'bbox_face_ids': self.face_ids}
            all_kwargs.update(kwargs)
            if extra_face_params is not None and face_id in extra_face_params:
                all_kwargs.update(**extra_face_params[face_id])
            self.mask_layouts[face_id]: MaskLayout = mask_layout_type(
                layout=layout,
                name=name,
                version=version,
                with_grid=with_grid,
                face_id=face_id,
                chips_map=chips_map if chips_map is not None else [[]],
                **all_kwargs
            )

    def add_chips_map(self, chips_map, **kwargs):
        for face_id in self.face_ids:
            self.mask_layouts[face_id].add_chips_map(chips_map, **kwargs)



# This code is part of KQCircuits
# Copyright (C) 2022 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import math

from kqcircuits.elements.element import Element
from kqcircuits.junctions.squid import Squid
from kqcircuits.junctions.manhattan import Manhattan
from kqcircuits.junctions.manhattan_single_junction import ManhattanSingleJunction
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.qubits.qubit import Qubit
from kqcircuits.pya_resolver import pya
from kqcircuits.util.refpoints import WaveguideToSimPort, JunctionSimPort


@add_parameters_from(Squid, junction_type="Manhattan")
@add_parameters_from(Manhattan)
@add_parameters_from(ManhattanSingleJunction)
class DoublePads(Qubit):
    """A two-island qubit, consisting of two rounded rectangles shunted by a junction, with one capacitive coupler.

    Contains a coupler on the north edge and two separate qubit islands in the center
    joined by a junction or SQUID loaded from another library.
    Refpoint for a readout line at the opening to the coupler and a modifiable refpoint for
    a driveline.
    """

    ground_gap = Param(pdt.TypeList, "Width, height of the ground gap (µm, µm)", [700, 700])
    ground_gap_r = Param(pdt.TypeDouble, "Ground gap rounding radius", 50, unit="μm")
    coupler_extent = Param(pdt.TypeList, "Width, height of the coupler (µm, µm)", [150, 20])
    coupler_r = Param(pdt.TypeDouble, "Coupler rounding radius", 10, unit="μm")
    coupler_a = Param(pdt.TypeDouble, "Width of the coupler waveguide center conductor", Element.a, unit="μm")
    coupler_offset = Param(pdt.TypeDouble, "Distance from first qubit island to coupler", 20, unit="μm")
    squid_offset = Param(pdt.TypeDouble, "Offset between SQUID center and qubit center", 0, unit="μm")
    island1_extent = Param(pdt.TypeList, "Width, height of the first qubit island (µm, µm)", [500, 100])
    island1_r = Param(pdt.TypeDouble, "First qubit island rounding radius", 50, unit="μm")
    island2_extent = Param(pdt.TypeList, "Width, height of the second qubit island (µm, µm)", [500, 100])
    island2_r = Param(pdt.TypeDouble, "Second qubit island rounding radius", 50, unit="μm")
    drive_position = Param(pdt.TypeList, "Coordinate for the drive port (µm, µm)", [-450, 0])
    island1_taper_width = Param(pdt.TypeDouble, "First qubit island tapering width on the island side", 10, unit="µm")
    island1_taper_junction_width = Param(pdt.TypeDouble,
                                         "First qubit island tapering width on the junction side", 10, unit="µm")
    island2_taper_width = Param(pdt.TypeDouble, "Second qubit island tapering width on the island side", 10, unit="µm")
    island2_taper_junction_width = Param(pdt.TypeDouble,
                                         "Second qubit island tapering width on the junction side", 10, unit="µm")

    island_island_gap = Param(pdt.TypeDouble, "Island to island gap distance", 70, unit="µm")

    def build(self):

        # Qubit base
        ground_gap_points = [
            pya.DPoint(float(self.ground_gap[0]) / 2,  float(self.ground_gap[1]) / 2),
            pya.DPoint(float(self.ground_gap[0]) / 2, -float(self.ground_gap[1]) / 2),
            pya.DPoint(-float(self.ground_gap[0]) / 2, -float(self.ground_gap[1]) / 2),
            pya.DPoint(-float(self.ground_gap[0]) / 2,  float(self.ground_gap[1]) / 2),
        ]
        ground_gap_polygon = pya.DPolygon(ground_gap_points)
        ground_gap_region = pya.Region(ground_gap_polygon.to_itype(self.layout.dbu))
        ground_gap_region.round_corners(self.ground_gap_r / self.layout.dbu,
                                        self.ground_gap_r / self.layout.dbu, self.n)

        # SQUID
        # Create temporary SQUID cell to calculate SQUID height
        temp_squid_cell = self.add_element(Squid, junction_type=self.junction_type)
        temp_squid_ref = self.get_refpoints(temp_squid_cell)
        squid_height = temp_squid_ref["port_common"].distance(pya.DPoint(0, 0))
        # Now actually add SQUID
        squid_transf = pya.DCplxTrans(1, 0, False, pya.DVector(0, self.squid_offset - squid_height / 2))
        self.produce_squid(squid_transf)

        taper_height = (self.island_island_gap - squid_height)/2

        # First island
        island1_region = self._build_island1(squid_height, taper_height)

        # Second island
        island2_region = self._build_island2(squid_height, taper_height)

        # Coupler gap
        coupler_region = self._build_coupler((self.squid_offset + squid_height / 2)
                                             + taper_height + float(self.island1_extent[1]))

        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(
            ground_gap_region - coupler_region - island1_region - island2_region
        )

        # Protection
        protection_polygon = pya.DPolygon([
            p + pya.DVector(
                    math.copysign(self.margin, p.x),
                    math.copysign(self.margin, p.y)
            ) for p in ground_gap_points
        ])
        protection_region = pya.Region(protection_polygon.to_itype(self.layout.dbu))
        protection_region.round_corners(
            (self.ground_gap_r + self.margin) / self.layout.dbu,
            (self.ground_gap_r + self.margin) / self.layout.dbu,
            self.n
        )
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(protection_region)

        # Coupler port
        self.add_port("cplr", pya.DPoint(0, float(self.ground_gap[1]) / 2),
                      direction=pya.DVector(pya.DPoint(0, float(self.ground_gap[1]))))

        # Drive port
        self.add_port("drive", pya.DPoint(float(self.drive_position[0]), float(self.drive_position[1])),
                      direction=pya.DVector(float(self.drive_position[0]), float(self.drive_position[1])))

    def _build_coupler(self, first_island_top_edge):
        coupler_top_edge = first_island_top_edge + self.coupler_offset + float(self.coupler_extent[1])
        coupler_polygon = pya.DPolygon([
            pya.DPoint(-float(self.coupler_extent[0]) / 2, coupler_top_edge),
            pya.DPoint(-float(self.coupler_extent[0]) / 2, first_island_top_edge + self.coupler_offset),
            pya.DPoint(float(self.coupler_extent[0]) / 2, first_island_top_edge + self.coupler_offset),
            pya.DPoint(float(self.coupler_extent[0]) / 2, coupler_top_edge),
        ])
        coupler_region = pya.Region(coupler_polygon.to_itype(self.layout.dbu))
        coupler_region.round_corners(self.coupler_r / self.layout.dbu, self.coupler_r / self.layout.dbu, self.n)
        coupler_path_polygon = pya.DPolygon([
            pya.DPoint(-self.coupler_a / 2, (float(self.ground_gap[1]) / 2)),
            pya.DPoint(self.coupler_a / 2, (float(self.ground_gap[1]) / 2)),
            pya.DPoint(self.coupler_a / 2, coupler_top_edge),
            pya.DPoint(-self.coupler_a / 2, coupler_top_edge),
        ])
        coupler_path = pya.Region(coupler_path_polygon.to_itype(self.layout.dbu))
        return coupler_region + coupler_path

    def _build_island1(self, squid_height, taper_height):
        island1_bottom = self.squid_offset + squid_height / 2
        island1_polygon = pya.DPolygon([
            pya.DPoint(-float(self.island1_extent[0]) / 2,
                       island1_bottom + taper_height + float(self.island1_extent[1])),
            pya.DPoint(float(self.island1_extent[0]) / 2,
                       island1_bottom + taper_height + float(self.island1_extent[1])),
            pya.DPoint(float(self.island1_extent[0]) / 2, island1_bottom + taper_height),
            pya.DPoint(-float(self.island1_extent[0]) / 2, island1_bottom + taper_height),
        ])
        island1_region = pya.Region(island1_polygon.to_itype(self.layout.dbu))
        island1_region.round_corners(self.island1_r / self.layout.dbu, self.island1_r / self.layout.dbu, self.n)
        island1_taper = pya.Region(pya.DPolygon([
            pya.DPoint(self.island1_taper_width / 2, island1_bottom + taper_height),
            pya.DPoint(self.island1_taper_junction_width / 2, island1_bottom),
            pya.DPoint(-self.island1_taper_junction_width / 2, island1_bottom),
            pya.DPoint(-self.island1_taper_width / 2, island1_bottom + taper_height),
        ]).to_itype(self.layout.dbu))

        return island1_region + island1_taper

    def _build_island2(self, squid_height, taper_height):
        island2_top = self.squid_offset - squid_height / 2
        island2_polygon = pya.DPolygon([
            pya.DPoint(-float(self.island2_extent[0]) / 2,
                       island2_top - taper_height - float(self.island2_extent[1])),
            pya.DPoint(float(self.island2_extent[0]) / 2,
                       island2_top - taper_height - float(self.island2_extent[1])),
            pya.DPoint(float(self.island2_extent[0]) / 2, island2_top - taper_height),
            pya.DPoint(-float(self.island2_extent[0]) / 2, island2_top - taper_height),
        ])
        island2_region = pya.Region(island2_polygon.to_itype(self.layout.dbu))
        island2_region.round_corners(self.island2_r / self.layout.dbu, self.island2_r / self.layout.dbu, self.n)
        island2_taper = pya.Region(pya.DPolygon([
            pya.DPoint(self.island2_taper_width / 2, island2_top - taper_height),
            pya.DPoint(self.island2_taper_junction_width / 2, island2_top),
            pya.DPoint(-self.island2_taper_junction_width / 2, island2_top),
            pya.DPoint(-self.island2_taper_width / 2, island2_top - taper_height),
        ]).to_itype(self.layout.dbu))
        return island2_region + island2_taper

    @classmethod
    def get_sim_ports(cls, simulation):
        return [JunctionSimPort(), WaveguideToSimPort("port_cplr", side="top")]



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
from math import sqrt
from kqcircuits.util.parameters import Param, pdt, add_parameters_from, pya
from kqcircuits.elements.element import Element

from kqcircuits.elements.waveguide_composite import WaveguideComposite
from kqcircuits.elements.waveguide_composite import Node
from kqcircuits.elements.finger_capacitor_square import FingerCapacitorSquare
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter
from kqcircuits.elements.finger_capacitor_taper import FingerCapacitorTaper
from kqcircuits.util.refpoints import RefpointToEdgePort

@add_parameters_from(FingerCapacitorSquare, finger_gap_end=3)
@add_parameters_from(FingerCapacitorTaper, finger_width=10, finger_number=2, finger_gap=3)
class CapacitiveXCoupler(Element):
    """
    Capacitive coupler for testing FEM computations.
    """
    x_coupler_length = Param(pdt.TypeDouble, "Length of Capacitive X Coupler", 500)
    x_coupler_height = Param(pdt.TypeDouble, "Height of Capacitive X Coupler", 500)
    x_coupler_variant = Param(pdt.TypeString, "Coupler variant, either (+) or (x)", "+", choices=["+", "x"])
    remove_capacitors = Param(pdt.TypeBoolean, "Remove capacitors from the X Coupler", False)


    def build(self):
        length = self.x_coupler_length
        height = self.x_coupler_height
        if length>height:
            center = height
            input_length = (length-center)/2.
        else:
            center = length
            input_length = (height-center)/2.

        p0 = (0,0)

        if self.x_coupler_variant == "x":
            if height > length:
                p11 = (-center/2., -input_length-length/2.)
                p12 = (-length/2., -center/2.)
                p13 = (-length/4., -center/4.)
            else:
                p11 = (-input_length-center/2., -height/2.)
                p12 = (-center/2., -height/2.)
                p13 = (-center/4., -height/4.)

            p21 = (-p11[0], p11[1])
            p22 = (-p12[0], p12[1])
            p23 = (-p13[0], p13[1])

            p31 = (-p11[0],-p11[1])
            p32 = (-p12[0],-p12[1])
            p33 = (-p13[0],-p13[1])

            p41 = ( p11[0],-p11[1])
            p42 = ( p12[0],-p12[1])
            p43 = ( p13[0],-p13[1])

            p02 = ((self.a/2+self.b)/sqrt(2), (-self.a/2-self.b)/sqrt(2))
            p04 = ((-self.a/2-self.b)/sqrt(2), (self.a/2+self.b)/sqrt(2))
        elif self.x_coupler_variant == "+":
            p11 = (0., -height/2.)
            p12 = (0., -height/3.)
            p13 = (0., -height/4.)

            p21 = (length/2., 0.)
            p22 = (length/3., 0.)
            p23 = (length/4., 0.)

            p31 = (0.,-p11[1])
            p32 = (0.,-p12[1])
            p33 = (0.,-p13[1])

            p41 = (-p21[0],0.)
            p42 = (-p22[0],0.)
            p43 = (-p23[0],0.)

            p02 = (self.a/2+self.b, 0.)
            p04 = (-self.a/2-self.b, 0.)

        self.refpoints['p11'] = pya.DPoint(*p11)
        self.refpoints['p21'] = pya.DPoint(*p21)
        self.refpoints['p31'] = pya.DPoint(*p31)
        self.refpoints['p41'] = pya.DPoint(*p41)

        def _add_capacitor_node(p):
            return Node(p, FingerCapacitorSquare,
                    finger_number=self.finger_number,
                    finger_width=self.finger_width,
                    finger_gap_end=self.finger_gap_end,
                    finger_gap=self.finger_gap,
                    a2=self.a,
                    b2=self.b,
                    a=self.a,
                    b=self.b,
                    ground_padding=self.b)

        splitter_port_length = self.a/2 + self.b

        nodes1 = [Node(p11),
                  Node(p12),
                  _add_capacitor_node(p13),
                  Node(p0, WaveguideCoplanarSplitter,
                        angles=[0, 180, 90, 270],
                        lengths=4*[splitter_port_length],
                        a=self.a,
                        b=self.b),
                  _add_capacitor_node(p33),
                  Node(p32),
                  Node(p31)]
        nodes2 = [Node(p41),
                  Node(p42),
                  _add_capacitor_node(p43),
                  Node(p04)]
        nodes3 = [Node(p21),
                  Node(p22),
                  _add_capacitor_node(p23),
                  Node(p02)]

        if self.remove_capacitors:
            nodes1.pop(4)
            nodes1.pop(2)
            nodes2.pop(2)
            nodes3.pop(2)

        if self.x_coupler_variant == 'x' and length == height:
            nodes1.pop(1)
            nodes1.pop(-1)
            nodes2.pop(1)
            nodes3.pop(1)

        self.insert_cell(WaveguideComposite, nodes=nodes1)
        self.insert_cell(WaveguideComposite, nodes=nodes2)
        self.insert_cell(WaveguideComposite, nodes=nodes3)

    @classmethod
    def get_sim_ports(cls, simulation):
        return [RefpointToEdgePort('p11'),
                RefpointToEdgePort('p21'),
                RefpointToEdgePort('p31'),
                RefpointToEdgePort('p41')]



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


import numpy

from kqcircuits.qubits.qubit import Qubit
from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.test_structures.test_structure import TestStructure
from kqcircuits.defaults import default_junction_test_pads_type
from kqcircuits.test_structures.junction_test_pads import junction_test_pads_type_choices
from kqcircuits.junctions.manhattan import Manhattan
from kqcircuits.junctions.junction import Junction


@add_parameters_from(Manhattan)
@add_parameters_from(Qubit, "junction_type", "junction_width", "loop_area", "mirror_squid",
                     "junction_parameters", "_junction_parameters")
class JunctionTestPads(TestStructure):
    """Base class for junction test structures."""

    default_type = default_junction_test_pads_type

    pad_width = Param(pdt.TypeDouble, "Pad width", 500, unit="μm")
    area_height = Param(pdt.TypeDouble, "Area height", 1900, unit="μm")
    area_width = Param(pdt.TypeDouble, "Area width", 1300, unit="μm")
    junctions_horizontal = Param(pdt.TypeBoolean, "Horizontal (True) or vertical (False) junctions", True)
    pad_spacing = Param(pdt.TypeDouble, "Spacing between different pad pairs", 100, unit="μm")
    only_pads = Param(pdt.TypeBoolean, "Only produce pads, no junctions", False)
    pad_configuration = Param(pdt.TypeString, "Pad configuration", "2-port", choices=["2-port", "4-port"])
    junction_width_steps = Param(pdt.TypeList, "Automatically generate junction widths [start, step]", [0, 0],
                                 unit="μm, μm")
    junction_widths = Param(pdt.TypeList, "Optional junction widths for individual junctions", [],
                            docstring="Override the junction widths with these values.")
    junction_test_pads_type = Param(pdt.TypeString, "Type of junction test pads", default_type,
                                    choices=junction_test_pads_type_choices)

    junction_test_pads_parameters = Param(pdt.TypeString, "Extra JunctionTestPads Parameters", "{}")
    _junction_test_pads_parameters = Param(pdt.TypeString, "Previous state of *_parameters", "{}", hidden=True)

    produce_squid = Qubit.produce_squid

    @classmethod
    def create(cls, layout, library=None, junction_test_pads_type=None, **parameters):
        """Create a JunctionTestPads cell in layout."""
        return cls.create_subtype(layout, library, junction_test_pads_type, **parameters)[0]

    def coerce_parameters_impl(self):
        self.sync_parameters(Junction)
        self.sync_parameters(JunctionTestPads)

    def _produce_impl(self):
        if self.pad_configuration == "2-port":
            self._produce_two_port_junction_tests()
        if self.pad_configuration == "4-port":
            self._produce_four_port_junction_tests()

    def _next_junction_width(self, idx):
        """Get the next junction width

        Try first the `junction_widths` list, if available, if not then generate it based on `start`
        and `step`, unless `step` is 0, in this case just use the default `junction_width`.
        """
        start, step = [float(x) for x in self.junction_width_steps]
        if idx < len(self.junction_widths) and self.junction_widths[idx] != '':
            return float(self.junction_widths[idx])
        elif not (start == 0 and step == 0):
            return start + idx * step
        return self.junction_width

    def _produce_two_port_junction_tests(self):

        pads_region = pya.Region()
        pad_step = self.pad_spacing + self.pad_width
        arm_width = 8

        junction_idx = 0
        y_flip = -1 if self.face_ids[0] == '2b1' else 1
        if self.junctions_horizontal:
            for x in numpy.arange(self.pad_spacing*1.5 + self.pad_width, self.area_width - pad_step, 2*pad_step,
                                  dtype=numpy.double):
                for y in numpy.arange(self.pad_spacing + self.pad_width*0.5, self.area_height - self.pad_width / 2,
                                      pad_step, dtype=numpy.double):
                    self.produce_pad(x - pad_step / 2, y, pads_region, self.pad_width, self.pad_width)
                    self.produce_pad(x + pad_step / 2, y, pads_region, self.pad_width, self.pad_width)
                    self._next_width = self._next_junction_width(junction_idx)
                    self._produce_junctions(x, y, pads_region, arm_width, junction_idx)
                    self.refpoints["probe_{}_l".format(junction_idx)] = pya.DPoint(x - pad_step * y_flip / 2, y)
                    self.refpoints["probe_{}_r".format(junction_idx)] = pya.DPoint(x + pad_step * y_flip/ 2, y)
                    junction_idx += 1
        else:
            for y in numpy.arange(self.pad_spacing*1.5 + self.pad_width, self.area_height - pad_step, 2*pad_step,
                                  dtype=numpy.double):
                for x in numpy.arange(self.pad_spacing + self.pad_width*0.5, self.area_width - self.pad_width / 2,
                                      pad_step, dtype=numpy.double):
                    self.produce_pad(x, y - pad_step / 2, pads_region, self.pad_width, self.pad_width)
                    self.produce_pad(x, y + pad_step / 2, pads_region, self.pad_width, self.pad_width)
                    self._next_width = self._next_junction_width(junction_idx)
                    self._produce_junctions(x, y, pads_region, arm_width, junction_idx)
                    self.refpoints["probe_{}_l".format(junction_idx)] = pya.DPoint(x, y - pad_step / 2)
                    self.refpoints["probe_{}_r".format(junction_idx)] = pya.DPoint(x, y + pad_step / 2)
                    junction_idx += 1

        self.produce_etched_region(pads_region, pya.DPoint(self.area_width / 2, self.area_height / 2), self.area_width,
                                   self.area_height)

    def _produce_junctions(self, x, y, pads_region, arm_width, index):

        if not self.only_pads:
            self._produce_squid_and_arms(x, y, pads_region, arm_width, index)

    def _produce_four_port_junction_tests(self):

        pads_region = pya.Region()
        junction_idx = 0
        step = 2 * (self.pad_width + self.pad_spacing)

        for x in numpy.arange(self.pad_spacing*1.5 + self.pad_width, self.area_width - step / 2, step,
                              dtype=numpy.double):
            for y in numpy.arange(self.pad_spacing*1.5 + self.pad_width, self.area_height - step / 2, step,
                                  dtype=numpy.double):

                if self.only_pads:
                    self.produce_four_point_pads(pads_region, self.pad_width, self.pad_width, self.pad_spacing,
                                                 self.pad_spacing, False, pya.DTrans(0, False, x, y),
                                                 "probe_{}".format(junction_idx))
                else:
                    self.produce_four_point_pads(pads_region, self.pad_width, self.pad_width, self.pad_spacing,
                                                 self.pad_spacing, True,
                                                 pya.DTrans(0 if self.junctions_horizontal else 1, False, x, y),
                                                 "probe_{}".format(junction_idx))
                    self._next_width = self._next_junction_width(junction_idx)
                    self._produce_junctions(x, y, pads_region, 5, junction_idx)

                junction_idx += 1

        self.produce_etched_region(pads_region, pya.DPoint(self.area_width / 2, self.area_height / 2), self.area_width,
                                   self.area_height)

    def _produce_squid_and_arms(self, x, y, pads_region, arm_width, index, only_arms=False):
        """Produces a squid and arms for connecting it to pads.

        The squid is inserted as a subcell. The arm shapes are inserted to pads_region, and their shape depends on
        arm_width and self.junctions_horizontal.

        Args:
           x: x-coordinate of squid origin
           y: y-coordinate of squid origin
           pads_region: Region to which the arm shapes are inserted
           arm_width: width of the arms
           only_arms: Boolean argument that allows to choose whether to create the arms and the squid device or
                            only the arms
        """

        arm_length = 11
        junction_spacing = self.junction_spacing

        # squid
        trans = pya.DCplxTrans(x, y - junction_spacing)
        if not self.junctions_horizontal:
            trans = pya.DCplxTrans(x - junction_spacing, y)
        squid_ref_rel = self.produce_squid(trans, only_arms=only_arms, junction_width=self._next_width,
                                           squid_index=index)
        if "right_side" in squid_ref_rel:
            arm_length = squid_ref_rel["right_side"].x - 0.5
        pos_rel_squid_top = squid_ref_rel["port_common"]

        if self.junctions_horizontal:
            # arm below
            arm1 = pya.DBox(
                pya.DPoint(x + arm_length, y - junction_spacing),
                pya.DPoint(x - self.pad_spacing / 2, y - arm_width - junction_spacing),
            )
            pads_region.insert(arm1.to_itype(self.layout.dbu))
            # arm above
            arm2 = pya.DBox(
                trans*pos_rel_squid_top + pya.DVector(-4, 0),
                trans*pos_rel_squid_top + pya.DVector(self.pad_spacing / 2, arm_width),
            )
            pads_region.insert(arm2.to_itype(self.layout.dbu))
        else:
            # arm below
            arm1 = pya.DBox(
                pya.DPoint(x + arm_length - junction_spacing, y),
                pya.DPoint(x - arm_length - junction_spacing, y - arm_width),
            )
            pads_region.insert(arm1.to_itype(self.layout.dbu))
            arm2 = pya.DBox(
                pya.DPoint(x + arm_width / 2 - junction_spacing, y),
                pya.DPoint(x - arm_width / 2 - junction_spacing, y - self.pad_spacing / 2),
            )
            pads_region.insert(arm2.to_itype(self.layout.dbu))
            # arm above
            arm3 = pya.DBox(
                trans*pos_rel_squid_top + pya.DVector(-arm_width / 2, 0),
                trans*pos_rel_squid_top + pya.DVector(arm_width / 2, self.pad_spacing / 2),
            )
            pads_region.insert(arm3.to_itype(self.layout.dbu))



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import logging

from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt

from kqcircuits.chips.chip import Chip
from kqcircuits.elements.waveguide_composite import WaveguideComposite, Node


class SampleHolderTest(Chip):
    """
    The PCell declaration for a SampleHolderTest chip.

    SampleHolderTest has parametrized launcher configuration (launcher dimensions and number of launchers).
    The launchers are connected pairwise by coplanar waveguides.
    """
    n_launchers = Param(pdt.TypeInt, "Number of launchers", 40, unit="")
    launcher_pitch = Param(pdt.TypeDouble, "Launcher pitch", 635, unit="μm")
    launcher_width = Param(pdt.TypeDouble, "Launcher width", 160, unit="μm")
    launcher_gap = Param(pdt.TypeDouble, "Launcher gap", 96, unit="μm")
    launcher_indent = Param(pdt.TypeDouble, "Launcher indent from edge", 520, unit="μm")

    def build(self):
        nr_pads_per_side = int(self.n_launchers / 4.)

        # North edge
        launcher_assigments = {i: f"PL-{i}-IN" for i in range(1, nr_pads_per_side + 1)}
        # East edge
        launcher_assigments.update({pin: f"PL-{pl+1}-OUT" for pl, pin in
                                    enumerate(range(2 * nr_pads_per_side, nr_pads_per_side, -1))})
        # South edge
        launcher_assigments.update({pin: f"PL-{pl + nr_pads_per_side + 1}-OUT" for pl, pin in
                                    enumerate(range(2 * nr_pads_per_side + 1, 3 * nr_pads_per_side + 1))})
        # West edge
        launcher_assigments.update({pin: f"PL-{pl + nr_pads_per_side + 1}-IN" for pl, pin in
                                    enumerate(range(4 * nr_pads_per_side, 3 * nr_pads_per_side, -1))})

        self.produce_n_launchers(self.n_launchers, "RF", self.launcher_width, self.launcher_gap, self.launcher_indent,
                                 self.launcher_pitch, launcher_assignments=launcher_assigments)

        def _produce_waveguide(i, straight_distance, first_port, second_port):
            cell = self.add_element(WaveguideComposite, nodes=[
                Node(self.refpoints[f'PL-{i}-{first_port}_port']),
                Node(self.refpoints[f'PL-{i}-{first_port}_port_corner'] + pya.DVector(0, straight_distance)),
                Node(self.refpoints[f'PL-{i}-{second_port}_port_corner'] + pya.DVector(straight_distance, 0)),
                Node(self.refpoints[f'PL-{i}-{second_port}_port']),
            ])
            self.insert_cell(cell)

            logging.info(f"{self.name_chip}: Waveguide PL-{i} length: {cell.length()}")

        for i in range(1, nr_pads_per_side + 1):
            _produce_waveguide(i, -1200, first_port="IN", second_port="OUT")

        for i in range(nr_pads_per_side + 1, nr_pads_per_side * 2 + 1):
            _produce_waveguide(i, 1200, first_port="OUT", second_port="IN")



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).

import os
import stat

import json
import logging

from distutils.dir_util import copy_tree
from pathlib import Path

from kqcircuits.util.export_helper import write_commit_reference_file
from kqcircuits.util.geometry_json_encoder import GeometryJsonEncoder
from kqcircuits.simulations.export.util import export_layers
from kqcircuits.defaults import ANSYS_EXECUTABLE, ANSYS_SCRIPT_PATHS
from kqcircuits.simulations.simulation import Simulation


def copy_ansys_scripts_to_directory(path: Path, import_script_folder='scripts'):
    """
    Copies Ansys scripts into directory path.

    Arguments:
        path: Location where to copy scripts folder.
        import_script_folder: Name of the folder in its new location.
    """
    if path.exists() and path.is_dir():
        for script_path in ANSYS_SCRIPT_PATHS:
            copy_tree(str(script_path), str(path.joinpath(import_script_folder)), update=1)


def export_ansys_json(simulation: Simulation, path: Path, ansys_tool='hfss',
                      frequency_units="GHz", frequency=5, max_delta_s=0.1, percent_error=1, percent_refinement=30,
                      maximum_passes=12, minimum_passes=1, minimum_converged_passes=1,
                      sweep_enabled=True, sweep_start=0, sweep_end=10, sweep_count=101, sweep_type='interpolating',
                      max_delta_f=0.1, n_modes=2, mesh_size=None, substrate_loss_tangent=0,
                      dielectric_surfaces=None, simulation_flags=None, ansys_project_template=None,
                      integrate_energies=False):
    r"""
    Export Ansys simulation into json and gds files.

    Arguments:
        simulation: The simulation to be exported.
        path: Location where to write json and gds files.
        ansys_tool: Determines whether to use HFSS ('hfss') or Q3D Extractor ('q3d').
        frequency_units: Units of frequency.
        frequency: Frequency for mesh refinement. To set up multifrequency analysis in HFSS use list of numbers.
        max_delta_s: Stopping criterion in HFSS simulation.
        percent_error: Stopping criterion in Q3D simulation.
        percent_refinement: Percentage of mesh refinement on each iteration.
        maximum_passes: Maximum number of iterations in simulation.
        minimum_passes: Minimum number of iterations in simulation.
        minimum_converged_passes: Determines how many iterations have to meet the stopping criterion to stop simulation.
        sweep_enabled: Determines if HFSS frequency sweep is enabled.
        sweep_start: The lowest frequency in the sweep.
        sweep_end: The highest frequency in the sweep.
        sweep_count: Number of frequencies in the sweep.
        sweep_type: choices are "interpolating", "discrete" or "fast"
        max_delta_f: Maximum allowed relative difference in eigenfrequency (%). Used when ``ansys_tool`` is *eigenmode*.
        n_modes: Number of eigenmodes to solve. Used when ``ansys_tool`` is 'pyepr'.
        mesh_size(dict): Dictionary to determine manual mesh refinement on layers. Set key as the layer name and
            value as the maximal mesh element length inside the layer.
        substrate_loss_tangent: Bulk loss tangent (:math:`\tan{\delta}`) material parameter. 0 is off.
        dielectric_surfaces: Material parameters for TLS interfaces, used in post-processing field calculations
            from the participation sheets. Default is None. Input is of the form::

                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                    'th': 4e-9,  # thickness
                    'eps_r': 10  # relative permittivity
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                    'th': 2e-9,
                    'eps_r': 10
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                    'th': 2e-9,
                    'eps_r': 10
                },
        simulation_flags: Optional export processing, given as list of strings
        ansys_project_template: path to the simulation template
        integrate_energies: Calculate energy integrals over each layer and save them into a file

    Returns:
         Path to exported json file.
    """
    if simulation is None or not isinstance(simulation, Simulation):
        raise ValueError("Cannot export without simulation")
    if simulation_flags is None:
        simulation_flags = []

    # collect data for .json file
    json_data = {
        'ansys_tool': ansys_tool,
        **simulation.get_simulation_data(),
        'analysis_setup': {
            'frequency_units': frequency_units,
            'frequency': frequency,
            'max_delta_s': max_delta_s,  # stopping criterion for HFSS
            'percent_error': percent_error,  # stopping criterion for Q3D
            'percent_refinement': percent_refinement,
            'maximum_passes': maximum_passes,
            'minimum_passes': minimum_passes,
            'minimum_converged_passes': minimum_converged_passes,
            'sweep_enabled': sweep_enabled,
            'sweep_start': sweep_start,
            'sweep_end': sweep_end,
            'sweep_count': sweep_count,
            'sweep_type': sweep_type,
            'max_delta_f': max_delta_f,
            'n_modes': n_modes,
        },
        'mesh_size': {} if mesh_size is None else mesh_size,
        'substrate_loss_tangent': substrate_loss_tangent,
        'dielectric_surfaces': dielectric_surfaces,
        'simulation_flags': simulation_flags,
        'integrate_energies': integrate_energies
    }

    if ansys_project_template is not None:
        json_data['ansys_project_template'] = ansys_project_template

    # write .json file
    json_filename = str(path.joinpath(simulation.name + '.json'))
    with open(json_filename, 'w') as fp:
        json.dump(json_data, fp, cls=GeometryJsonEncoder, indent=4)

    # write .gds file
    gds_filename = str(path.joinpath(simulation.name + '.gds'))
    export_layers(gds_filename, simulation.layout, [simulation.cell],
                  output_format='GDS2',
                  layers=simulation.get_layers()
                  )

    return json_filename


def export_ansys_bat(json_filenames, path: Path, file_prefix='simulation', exit_after_run=False,
                     import_script_folder='scripts', import_script='import_and_simulate.py',
                     post_process_script='export_batch_results.py', intermediate_processing_command=None,
                     use_rel_path=True):
    """
    Create a batch file for running one or more already exported simulations.

    Arguments:
        json_filenames: List of paths to json files to be included into the batch.
        path: Location where to write the bat file.
        file_prefix: Name of the batch file to be created.
        exit_after_run: Defines if the Ansys Electronics Desktop is automatically closed after running the script.
        import_script_folder: Path to the Ansys-scripts folder.
        import_script: Name of import script file.
        post_process_script: Name of post processing script file.
        intermediate_processing_command: Command for intermediate steps between simulations.
            Default is None, which doesn't enable any processing. An example argument is ``python scripts/script.py``,
            which runs in the `.bat` as::

                python scripts/script.py json_filename.json

        use_rel_path: Determines if to use relative paths.

    Returns:
         Path to exported bat file.
    """
    run_cmd = 'RunScriptAndExit' if exit_after_run else 'RunScript'

    bat_filename = str(path.joinpath(file_prefix + '.bat'))
    with open(bat_filename, 'w') as file:
        file.write(
            '@echo off\n'\
            r'powershell -Command "Get-Process | Where-Object {$_.MainWindowTitle -like \"Run Simulations*\"} '\
                '| Select -ExpandProperty Id | Export-Clixml -path blocking_pids.xml"\n'\
            'title Run Simulations\n'\
            'powershell -Command "$sim_pids = Import-Clixml -Path blocking_pids.xml; if ($sim_pids) '\
                r'{ echo \"Waiting for $sim_pids\"; Wait-Process $sim_pids -ErrorAction SilentlyContinue }; '\
                'Remove-Item blocking_pids.xml"\n'
        )

        # Commands for each simulation
        for i, json_filename in enumerate(json_filenames):
            printing = 'echo Simulation {}/{} - {}\n'.format(
                i+1,
                len(json_filenames),
                str(Path(json_filename).relative_to(path)))
            file.write(printing)
            command = '"{}" -scriptargs "{}" -{} "{}"\n'.format(
                ANSYS_EXECUTABLE,
                str(Path(json_filename).relative_to(path) if use_rel_path else json_filename),
                run_cmd,
                str(Path(import_script_folder).joinpath(import_script)))
            file.write(command)
            # Possible processing between simulations
            if intermediate_processing_command is not None:
                command = '{} "{}"\n'.format(
                    intermediate_processing_command,
                    str(Path(json_filename).relative_to(path))
                )
                file.write(command)

        # Post-process command
        command = '"{}" -{} "{}"\n'.format(
            ANSYS_EXECUTABLE,
            run_cmd,
            str(Path(import_script_folder).joinpath(post_process_script)))
        file.write(command)

    # Make the bat file executable in linux
    os.chmod(bat_filename, os.stat(bat_filename).st_mode | stat.S_IEXEC)

    return bat_filename


def export_ansys(simulations, path: Path, ansys_tool='hfss', import_script_folder='scripts', file_prefix='simulation',
                 frequency_units="GHz", frequency=5, max_delta_s=0.1, percent_error=1, percent_refinement=30,
                 maximum_passes=12, minimum_passes=1, minimum_converged_passes=1,
                 sweep_enabled=True, sweep_start=0, sweep_end=10, sweep_count=101, sweep_type='interpolating',
                 max_delta_f=0.1, n_modes=2, mesh_size=None, substrate_loss_tangent=0,
                 dielectric_surfaces=None, exit_after_run=False,
                 import_script='import_and_simulate.py', post_process_script='export_batch_results.py',
                 intermediate_processing_command=None, use_rel_path=True, simulation_flags=None,
                 ansys_project_template=None, integrate_energies=False, skip_errors=False):
    r"""
    Export Ansys simulations by writing necessary scripts and json, gds, and bat files.

    Arguments:
        simulations: List of simulations to be exported.
        path: Location where to write export files.
        ansys_tool: Determines whether to use HFSS ('hfss'), Q3D Extractor ('q3d') or HFSS eigenmode ('eigenmode').
        import_script_folder: Path to the Ansys-scripts folder.
        file_prefix: Name of the batch file to be created.
        frequency_units: Units of frequency.
        frequency: Frequency for mesh refinement. To set up multifrequency analysis in HFSS use list of numbers.
        max_delta_s: Stopping criterion in HFSS simulation.
        percent_error: Stopping criterion in Q3D simulation.
        percent_refinement: Percentage of mesh refinement on each iteration.
        maximum_passes: Maximum number of iterations in simulation.
        minimum_passes: Minimum number of iterations in simulation.
        minimum_converged_passes: Determines how many iterations have to meet the stopping criterion to stop simulation.
        sweep_enabled: Determines if HFSS frequency sweep is enabled.
        sweep_start: The lowest frequency in the sweep.
        sweep_end: The highest frequency in the sweep.
        sweep_count: Number of frequencies in the sweep.
        sweep_type: choices are "interpolating", "discrete" or "fast"
        max_delta_f: Maximum allowed relative difference in eigenfrequency (%). Used when ``ansys_tool`` is *eigenmode*.
        n_modes: Number of eigenmodes to solve. Used when ``ansys_tool`` is 'eigenmode'.
        mesh_size(dict): Dictionary to determine manual mesh refinement on layers. Set key as the layer name and
            value as the maximal mesh element length inside the layer.
        substrate_loss_tangent: Bulk loss tangent (:math:`\tan{\delta}`) material parameter. 0 is off.
        dielectric_surfaces: Material parameters for TLS interfaces, used in post-processing field calculations
            from the participation sheets. Default is None. Input is of the form::

                'layerMA': {  # metal–vacuum
                    'tan_delta_surf': 0.001,  # loss tangent
                    'th': 4e-9,  # thickness
                    'eps_r': 10  # relative permittivity
                },
                'layerMS': { # metal–substrate
                    'tan_delta_surf': 0.001,
                    'th': 2e-9,
                    'eps_r': 10
                },
                'layerSA': { # substrate–vacuum
                    'tan_delta_surf': 0.001,
                    'th': 2e-9,
                    'eps_r': 10
                },

        exit_after_run: Defines if the Ansys Electronics Desktop is automatically closed after running the script.
        import_script: Name of import script file.
        post_process_script: Name of post processing script file.
        intermediate_processing_command: Command for intermediate steps between simulations.
            Default is None, which doesn't enable any processing. An example argument is ``python scripts/script.py``,
            which runs in the `.bat` as::

                python scripts/script.py json_filename.json

        use_rel_path: Determines if to use relative paths.
        simulation_flags: Optional export processing, given as list of strings. See Simulation Export in docs.
        ansys_project_template: path to the simulation template
        integrate_energies: Calculate energy integrals over each layer and save them into a file
        skip_errors: Skip simulations that cause errors. Default is False.

            .. warning::

               **Use this carefully**, some of your simulations might not make sense physically and
               you might end up wasting time on bad simulations.

    Returns:
        Path to exported bat file.
    """
    write_commit_reference_file(path)
    copy_ansys_scripts_to_directory(path, import_script_folder=import_script_folder)
    json_filenames = []
    for simulation in simulations:
        try:
            json_filenames.append(export_ansys_json(simulation, path, ansys_tool=ansys_tool,
                                            frequency_units=frequency_units, frequency=frequency,
                                            max_delta_s=max_delta_s, percent_error=percent_error,
                                            percent_refinement=percent_refinement,
                                            maximum_passes=maximum_passes, minimum_passes=minimum_passes,
                                            minimum_converged_passes=minimum_converged_passes,
                                            sweep_enabled=sweep_enabled, sweep_start=sweep_start,
                                            sweep_end=sweep_end, sweep_count=sweep_count, sweep_type=sweep_type,
                                            max_delta_f=max_delta_f, n_modes=n_modes,
                                            mesh_size=mesh_size,
                                            substrate_loss_tangent=substrate_loss_tangent,
                                            dielectric_surfaces=dielectric_surfaces,
                                            simulation_flags=simulation_flags,
                                            ansys_project_template=ansys_project_template,
                                            integrate_energies=integrate_energies))
        except (IndexError, ValueError, Exception) as e:  # pylint: disable=broad-except
            if skip_errors:
                logging.warning(
                    f'Simulation {simulation.name} skipped due to {e.args}. '\
                    'Some of your other simulations might not make sense geometrically. '\
                    'Disable `skip_errors` to see the full traceback.'
                )
            else:
                raise UserWarning(
                    'Generating simulation failed. You can discard the errors using `skip_errors` in `export_ansys`. '\
                    'Moreover, `skip_errors` enables visual inspection of failed and successful simulation '\
                    'geometry files.'
                ) from e

    return export_ansys_bat(json_filenames, path, file_prefix=file_prefix, exit_after_run=exit_after_run,
                            import_script_folder=import_script_folder,
                            import_script=import_script, post_process_script=post_process_script,
                            intermediate_processing_command=intermediate_processing_command,
                            use_rel_path=use_rel_path)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).


from kqcircuits.pya_resolver import pya
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.test_structures.test_structure import TestStructure
from kqcircuits.elements.tsvs.tsv import Tsv
from kqcircuits.elements.tsvs.tsv_ellipse import TsvEllipse


@add_parameters_from(TsvEllipse, "*", tsv_diameter=10)
class TsvTestPattern(TestStructure):
    """PCell declaration for TSV test structures which resembles a TSV fencing for a CPW transmission line.

    Contains a given number of TSVs in `through silicon via`-layer with given vertical, horizontal pitch,
    placeholder for CPW, diameter and the lateral profile of the TSV.
    """

    cpw_distance = Param(pdt.TypeDouble, "CPW Placeholder distance", 100, unit="μm")
    hor_distance = Param(pdt.TypeDouble, "Horizontal pitch on TSV", 200, unit="μm")
    ver_distance = Param(pdt.TypeDouble, "Vertical pitch on TSV", 500, unit="μm")
    tsv_array_form = Param(pdt.TypeList, "TSV test layout", [2, 6, 6, 2, 6, 6, 2])

    def build(self):
        tsv_unit = self.add_element(Tsv)
        for i, ind in enumerate(self.tsv_array_form):
            index_at_cpw_pos = int(ind) / 2
            for j in range(int(ind)):
                if j >= index_at_cpw_pos:
                    y_pos = (j - (index_at_cpw_pos - 1 / 2)) * self.hor_distance + self.cpw_distance / 2 + 100
                else:
                    y_pos = (j - (index_at_cpw_pos - 1 / 2)) * self.hor_distance - self.cpw_distance / 2 - 100
                x_pos = (i - (len(self.tsv_array_form) - 1) / 2) * self.ver_distance
                trans = pya.DTrans(x_pos, y_pos)
                self.insert_cell(tsv_unit, trans)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/developers/osstmpolicy). IQM welcomes contributions to the code. Please see our contribution agreements
# for individuals (meetiqm.com/developers/clas/individual) and organizations (meetiqm.com/developers/clas/organization).
import importlib.metadata
import json
import logging
import subprocess
import platform
import sys
import os
import argparse
from sys import argv
from pathlib import Path

from kqcircuits.elements.element import get_refpoints
from kqcircuits.defaults import default_layers, TMP_PATH, STARTUPINFO, default_probe_types, default_probe_suffixes, \
    VERSION_PATHS, default_drc_runset, DRC_PATH
from kqcircuits.klayout_view import KLayoutView, MissingUILibraryException
from kqcircuits.pya_resolver import pya, is_standalone_session, klayout_executable_command


def generate_probepoints_json(cell, face='1t1'):
    # make autoprober json string for cell with reference points with magical names
    if cell is None or face not in ['1t1', '2b1']:
        error_text = f"Invalid face '{face}' or 'nil' cell ."
        error = ValueError(error_text)
        logging.exception(error_text, exc_info=error)
        raise error

    layout = cell.layout()

    refpoints = get_refpoints(layout.layer(default_layers["refpoints"]), cell)

    # Check existence of standard markers important for us
    if f"{face}_marker_nw" in refpoints and f"{face}_marker_se" in refpoints:
        markers = {'NW': refpoints[f"{face}_marker_nw"], 'SE': refpoints[f"{face}_marker_se"]}
    else:
        logging.error(f"There are no usable markers in {face}-face of the cell! Not a Chip?")
        return {}

    # flip top-markers back to top side
    if face == '2b1':
        origin = refpoints["1t1_marker_se"]
        markers = {k: flip(v, origin) for k, v in markers.items()}

    eu = 1e-3  # export unit

    # initialize dictionaries for each probe point group
    groups = {}
    for probe_name in default_probe_types.values():
        for marker_name, marker in markers.items():
            groups[f"{probe_name} {marker_name}"] = {
                "alignment": {"x": round(marker.x * eu, 3), "y": round(marker.y * eu, 3)},
                "pads": []
            }

    # divide probe points into groups by closest marker
    for probepoint_name, probepoint in refpoints.items():
        name_type = probepoint_name.split("_")[0]
        # if name_type starts with some probe_type, truncate name_type to be the probe_type
        for probe_type in default_probe_types:
            if name_type.startswith(probe_type):
                name_type = probe_type
                break
        # does the name correspond to a probepoint?
        if name_type in default_probe_types.keys() and probepoint_name.endswith(default_probe_suffixes):

            if face == '2b1':
                probepoint = flip(probepoint, origin)

            best_distance = 1e99
            closest_marker = None
            for marker, refpoint in markers.items():
                if refpoint.distance(probepoint) < best_distance:
                    best_distance = refpoint.distance(probepoint)
                    closest_marker = marker

            groups[f"{default_probe_types[name_type]} {closest_marker}"]["pads"].append({
                "id": probepoint_name,
                "x": round(probepoint.x * eu, 3),
                "y": round(probepoint.y * eu, 3),
            })

    # remove empty groups
    groups = {k: v for k, v in groups.items() if v["pads"]}

    # sort from left to right, bottom to top, for faster probing
    for group in groups.values():
        group["pads"] = sorted(group["pads"], key=lambda k: (k['x'], k['y']))

    # define JSON format
    comp_dict = {
        "groups": [{"id": name, **group} for name, group in groups.items()]
    }

    comp_json = json.dumps(comp_dict, indent=2, sort_keys=True)

    return comp_json

def flip(point, origin=pya.DPoint(0,0)):
    """Gets correct flip chip coordinates by setting a new origin and mirroring ``point`` by the y-axis."""
    return pya.DPoint(origin.x - point.x, point.y - origin.y)

def create_or_empty_tmp_directory(dir_name):
    """Creates directory into TMP_PATH or removes its content if it exists.
    Returns directory path.
    """
    def remove_content(path):
        """ Removes content of the directory path without removing directory itself."""
        for child in path.iterdir():
            if child.is_dir():
                remove_content(child)
                child.rmdir()
            else:
                child.unlink()

    parser = argparse.ArgumentParser()

    parser.add_argument("--simulation-export-path", type=str, default=None)
    args, _ = parser.parse_known_args()

    if args.simulation_export_path is not None:
        dir_path=Path(args.simulation_export_path)
    else:
        dir_path = TMP_PATH.joinpath(dir_name)

    if dir_path.exists() and dir_path.is_dir():
        remove_content(dir_path)
    else:
        dir_path.mkdir()

    return dir_path


def get_active_or_new_layout():
    """Tries to return active layout in GUI or returns new layout when running standalone."""
    try:
        klayoutview = KLayoutView(current=True)
        klayoutview.add_default_layers()
        return klayoutview.layout
    except MissingUILibraryException:
        return pya.Layout()


def write_commit_reference_file(path: Path, write_versions_file=True):
    """
    Writes file COMMIT_REFERENCE into given file path. The file includes current git revision number.
    If git repository is not found in given path, no file is written.
    """
    try:
        with open(path.joinpath('COMMIT_REFERENCE'), 'w') as file:
            for item in VERSION_PATHS.items():
                output = subprocess.check_output(['git', 'rev-parse', 'HEAD'], stderr=subprocess.DEVNULL,
                                             cwd=item[1], startupinfo=STARTUPINFO)
                file.write("{} revision number: {}".format(item[0], output.decode('ascii')))

    except subprocess.CalledProcessError:
        return

    if write_versions_file:
        write_export_machine_versions_file(path)

def write_export_machine_versions_file(path: Path):
    """
    Writes file EXPORT_MACHINE_VERSIONS into given file path.
    """
    versions = {}
    versions['platform'] = platform.platform()
    versions['python'] = sys.version_info
    versions['klayout'] = get_klayout_version()

    with open(path.joinpath('EXPORT_MACHINE_VERSIONS.json'), 'w') as file:
        json.dump(versions, file)

def open_with_klayout_or_default_application(filepath):
    """
    Tries to open file with KLayout. If KLayout is not found, opens file with operating system's default application.
    Implementation supports Windows, macOS, and Linux.
    """
    if argv[-1] == "-q":  # quiet mode, do not run viewer
        return

    exe = klayout_executable_command()
    if not exe:
        logging.warning("KLayout executable not found.")
    else:
        subprocess.call((exe, filepath))

def get_klayout_version():
    if is_standalone_session():
        return f"KLayout {importlib.metadata.version('klayout')}"
    else:
        return pya.Application.instance().version()


def export_drc_report(name, path, drc_script=default_drc_runset):
    """Run a DRC script on ``path/name.oas`` and export results in ``path/name_drc_report.lyrdb``."""

    drc_runset_path = os.path.join(DRC_PATH, drc_script)
    input_file = os.path.join(path, f"{name}.oas")
    output_file = os.path.join(path, f"{name}_drc_report.lyrdb")
    logging.info("Exporting DRC report to %s", output_file)

    try:
        subprocess.run([klayout_executable_command(), "-b", "-i",
                        "-r", drc_runset_path,
                        "-rd", f"output={output_file}",
                        input_file
                        ], check=True, startupinfo=STARTUPINFO)
    except subprocess.CalledProcessError as e:
        logging.error(e.output)
